from __future__ import annotations

import asyncio
import logging
import os
from contextlib import suppress
from pathlib import PurePath
from typing import Any, Awaitable, Dict, Final, Iterator, List, Optional, Tuple, Union

from reactivex.scheduler import ThreadPoolScheduler

from ..bili.live import Live
from ..core import Recorder, RecorderEventListener
from ..event.event_emitter import EventEmitter, EventListener
from ..exception import exception_callback, submit_exception
from ..flv.helpers import is_valid_flv_file
from ..flv.metadata_analysis import analyse_metadata
from ..flv.metadata_injection import InjectingProgress, inject_metadata
from ..logging.room_id import aio_task_with_room_id
from ..path import danmaku_path, extra_metadata_path, record_metadata_path
from ..utils.mixins import AsyncCooperationMixin, AsyncStoppableMixin, SupportDebugMixin
from .ffmpeg_metadata import make_metadata_file
from .helpers import copy_files_related, discard_dir, discard_file, get_extra_metadata
from .models import DeleteStrategy, PostprocessorStatus
from .remux import RemuxingProgress, RemuxingResult, remux_video
from .typing import Progress

__all__ = (
    'Postprocessor',
    'PostprocessorEventListener',
    'PostprocessorStatus',
    'DeleteStrategy',
)


logger = logging.getLogger(__name__)


class PostprocessorEventListener(EventListener):
    async def on_video_postprocessing_completed(
        self, postprocessor: Postprocessor, path: str
    ) -> None:
        ...


class Postprocessor(
    EventEmitter[PostprocessorEventListener],
    RecorderEventListener,
    AsyncStoppableMixin,
    AsyncCooperationMixin,
    SupportDebugMixin,
):
    _worker_semaphore: Final = asyncio.Semaphore(value=1)

    def __init__(
        self,
        live: Live,
        recorder: Recorder,
        *,
        remux_to_mp4: bool = False,
        inject_extra_metadata: bool = False,
        delete_source: DeleteStrategy = DeleteStrategy.AUTO,
    ) -> None:
        super().__init__()
        self._init_for_debug(live.room_id)

        self._live = live
        self._recorder = recorder

        self.remux_to_mp4 = remux_to_mp4
        self.inject_extra_metadata = inject_extra_metadata
        self.delete_source = delete_source

        self._status = PostprocessorStatus.WAITING
        self._postprocessing_path: Optional[str] = None
        self._postprocessing_progress: Optional[Progress] = None
        self._completed_files: List[str] = []

    @property
    def recorder(self) -> Recorder:
        return self._recorder

    @property
    def status(self) -> PostprocessorStatus:
        return self._status

    @property
    def postprocessing_path(self) -> Optional[str]:
        return self._postprocessing_path

    @property
    def postprocessing_progress(self) -> Optional[Progress]:
        return self._postprocessing_progress

    def get_completed_files(self) -> Iterator[str]:
        yield from iter(self._completed_files)

    async def on_recording_started(self, recorder: Recorder) -> None:
        # clear completed files of previous recording
        self._completed_files.clear()

    async def on_video_file_completed(self, recorder: Recorder, path: str) -> None:
        self._queue.put_nowait(path)

    async def on_danmaku_file_completed(self, recorder: Recorder, path: str) -> None:
        self._completed_files.append(path)

    async def _do_start(self) -> None:
        self._recorder.add_listener(self)

        self._queue: asyncio.Queue[str] = asyncio.Queue()
        self._scheduler = ThreadPoolScheduler()
        self._task = asyncio.create_task(self._worker())
        self._task.add_done_callback(exception_callback)

        logger.debug('Started postprocessor')

    async def _do_stop(self) -> None:
        self._recorder.remove_listener(self)

        await self._queue.join()
        self._task.cancel()
        with suppress(asyncio.CancelledError):
            await self._task

        del self._queue
        del self._scheduler
        del self._task

        logger.debug('Stopped postprocessor')

    @aio_task_with_room_id
    async def _worker(self) -> None:

        while True:
            self._status = PostprocessorStatus.WAITING
            self._postprocessing_path = None
            self._postprocessing_progress = None

            video_path = await self._queue.get()

            async with self._worker_semaphore:
                logger.debug(f'Postprocessing... {video_path}')

                await self._wait_for_metadata_file(video_path)

                try:
                    if video_path.endswith('.flv'):
                        if not await self._is_vaild_flv_file(video_path):
                            logger.warning(f'The flv file may be invalid: {video_path}')
                        if self.remux_to_mp4:
                            self._status = PostprocessorStatus.REMUXING
                            (
                                result_path,
                                remuxing_result,
                            ) = await self._remux_video_to_mp4(video_path)
                            if not self._debug:
                                await discard_file(
                                    extra_metadata_path(video_path), 'DEBUG'
                                )
                                if self._should_delete_source_files(remuxing_result):
                                    await discard_file(video_path)
                        elif self.inject_extra_metadata:
                            self._status = PostprocessorStatus.INJECTING
                            result_path = await self._inject_extra_metadata(video_path)
                        else:
                            result_path = video_path

                    elif video_path.endswith('.m3u8'):
                        if self.remux_to_mp4:
                            self._status = PostprocessorStatus.REMUXING
                            (
                                result_path,
                                remuxing_result,
                            ) = await self._remux_video_to_mp4(video_path)
                            await copy_files_related(video_path)
                            if result_path != video_path:
                                self._completed_files.append(danmaku_path(result_path))
                                self._completed_files.remove(danmaku_path(video_path))
                            if not self._debug:
                                if self._should_delete_source_files(remuxing_result):
                                    await discard_dir(os.path.dirname(video_path))
                        else:
                            result_path = video_path
                    else:
                        result_path = video_path

                    self._completed_files.append(result_path)
                    await self._emit(
                        'video_postprocessing_completed', self, result_path
                    )
                except Exception as exc:
                    submit_exception(exc)
                finally:
                    self._queue.task_done()

    async def _inject_extra_metadata(self, path: str) -> str:
        logger.info(f"Injecting metadata for '{path}' ...")
        try:
            try:
                metadata = await get_extra_metadata(path)
            except Exception as e:
                logger.warning(f'Failed to get extra metadata: {repr(e)}')
                logger.info(f"Analysing metadata for '{path}' ...")
                await self._analyse_metadata(path)
                metadata = await get_extra_metadata(path)
            else:
                if 'keyframes' not in metadata:
                    logger.warning('The keyframes metadata lost')
                    logger.info(f"Analysing metadata for '{path}' ...")
                    await self._analyse_metadata(path)
                    new_metadata = await get_extra_metadata(path)
                    metadata.update(new_metadata)
            await self._inject_metadata(path, metadata)
        except Exception as e:
            logger.error(f"Failed to inject metadata for '{path}': {repr(e)}")
            submit_exception(e)
        else:
            logger.info(f"Successfully injected metadata for '{path}'")
        return path

    async def _remux_video_to_mp4(self, in_path: str) -> Tuple[str, RemuxingResult]:
        if in_path.endswith('.flv'):
            out_path = str(PurePath(in_path).with_suffix('.mp4'))
            metadata_path = await make_metadata_file(in_path)
        elif in_path.endswith('.m3u8'):
            out_path = str(PurePath(in_path).parent.with_suffix('.mp4'))
            metadata_path = await make_metadata_file(in_path)
        else:
            raise NotImplementedError(in_path)

        logger.info(f"Remuxing '{in_path}' to '{out_path}' ...")
        remux_result = await self._remux_video(in_path, out_path, metadata_path)

        if remux_result.is_failed():
            logger.error(f"Failed to remux '{in_path}' to '{out_path}'")
            result_path = in_path
        elif remux_result.is_warned():
            logger.warning('Remuxing done, but ran into problems.')
            result_path = out_path
        elif remux_result.is_successful():
            logger.info(f"Successfully remuxed '{in_path}' to '{out_path}'")
            result_path = out_path
        else:
            pass

        logger.debug(f'ffmpeg output:\n{remux_result.output}')

        if not self._debug and in_path.endswith('.flv'):
            await discard_file(metadata_path, 'DEBUG')

        return result_path, remux_result

    def _analyse_metadata(self, path: str) -> Awaitable[None]:
        future: asyncio.Future[None] = asyncio.Future()
        self._postprocessing_path = path

        subscription = analyse_metadata(path, show_progress=True).subscribe(
            on_error=lambda e: future.set_exception(e),
            on_completed=lambda: future.set_result(None),
            scheduler=self._scheduler,
        )
        future.add_done_callback(lambda f: subscription.dispose())

        return future

    def _inject_metadata(self, path: str, metadata: Dict[str, Any]) -> Awaitable[None]:
        future: asyncio.Future[None] = asyncio.Future()
        self._postprocessing_path = path

        def on_next(value: InjectingProgress) -> None:
            self._postprocessing_progress = value

        subscription = inject_metadata(path, metadata, show_progress=True).subscribe(
            on_next=on_next,
            on_error=lambda e: future.set_exception(e),
            on_completed=lambda: future.set_result(None),
            scheduler=self._scheduler,
        )
        future.add_done_callback(lambda f: subscription.dispose())

        return future

    def _remux_video(
        self, in_path: str, out_path: str, metadata_path: str
    ) -> Awaitable[RemuxingResult]:
        future: asyncio.Future[RemuxingResult] = asyncio.Future()
        self._postprocessing_path = in_path

        def on_next(value: Union[RemuxingProgress, RemuxingResult]) -> None:
            if isinstance(value, RemuxingProgress):
                self._postprocessing_progress = value
            elif isinstance(value, RemuxingResult):
                future.set_result(value)

        subscription = remux_video(
            in_path,
            out_path,
            metadata_path,
            show_progress=True,
            remove_filler_data=True,
        ).subscribe(
            on_next=on_next,
            on_error=lambda e: future.set_exception(e),
            scheduler=self._scheduler,
        )
        future.add_done_callback(lambda f: subscription.dispose())

        return future

    async def _is_vaild_flv_file(self, video_path: str) -> bool:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, is_valid_flv_file, video_path)

    def _should_delete_source_files(self, remux_result: RemuxingResult) -> bool:
        if self.delete_source == DeleteStrategy.AUTO:
            if not remux_result.is_failed():
                return True
        elif self.delete_source == DeleteStrategy.SAFE:
            if not remux_result.is_failed() and not remux_result.is_warned():
                return True
        elif self.delete_source == DeleteStrategy.NEVER:
            return False

        return False

    async def _wait_for_metadata_file(self, video_path: str) -> None:
        loop = asyncio.get_running_loop()

        if video_path.endswith('.flv'):
            path = extra_metadata_path(video_path)
        elif video_path.endswith('.m3u8'):
            path = record_metadata_path(video_path)
        else:
            return

        for _ in range(10):
            if await loop.run_in_executor(None, os.path.isfile, path):
                break
            else:
                logger.debug(f'Not found metadata file: {path}')
                await asyncio.sleep(1)
        else:
            logger.warning(f'No such metadata file: {path}')



import os
import logging
import asyncio
from typing import Iterator, List, Optional

import attr
import psutil

from . import __prog__, __version__
from .flv.operators import MetaData, StreamProfile
from .disk_space import SpaceMonitor, SpaceReclaimer
from .bili.helpers import ensure_room_id
from .task import (
    RecordTaskManager,
    TaskData,
    TaskParam,
    VideoFileDetail,
    DanmakuFileDetail,
)
from .exception import ExistsError, ExceptionHandler, exception_callback
from .event.event_submitters import SpaceEventSubmitter
from .setting import (
    SettingsManager,
    Settings,
    SettingsIn,
    SettingsOut,
    TaskOptions,
)
from .setting.typing import KeySetOfSettings
from .notification import (
    EmailNotifier,
    ServerchanNotifier,
    PushdeerNotifier,
    PushplusNotifier,
    TelegramNotifier,
)
from .webhook import WebHookEmitter


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class AppInfo:
    name: str
    version: str
    pid: int
    ppid: int
    create_time: float
    cwd: str
    exe: str
    cmdline: List[str]


@attr.s(auto_attribs=True, slots=True, frozen=True)
class AppStatus:
    cpu_percent: float
    memory_percent: float
    num_threads: int


class Application:
    def __init__(self, settings: Settings) -> None:
        self._out_dir = settings.output.out_dir
        self._settings_manager = SettingsManager(self, settings)
        self._task_manager = RecordTaskManager(self._settings_manager)

    @property
    def info(self) -> AppInfo:
        p = psutil.Process(os.getpid())
        with p.oneshot():
            return AppInfo(
                name=__prog__,
                version=__version__,
                pid=p.pid,
                ppid=p.ppid(),
                create_time=p.create_time(),
                cwd=p.cwd(),
                exe=p.exe(),
                cmdline=p.cmdline(),
            )

    @property
    def status(self) -> AppStatus:
        p = psutil.Process(os.getpid())
        with p.oneshot():
            return AppStatus(
                cpu_percent=p.cpu_percent(),
                memory_percent=p.memory_percent(),
                num_threads=p.num_threads(),
            )

    def run(self) -> None:
        asyncio.run(self._run())

    async def _run(self) -> None:
        self._loop = asyncio.get_running_loop()

        await self.launch()
        try:
            self._interrupt_event = asyncio.Event()
            await self._interrupt_event.wait()
        finally:
            await self.exit()

    async def launch(self) -> None:
        self._setup()
        logger.debug(f'Default umask {os.umask(000)}')
        logger.info(f'Launched Application v{__version__}')
        task = asyncio.create_task(self._task_manager.load_all_tasks())
        task.add_done_callback(exception_callback)

    async def exit(self) -> None:
        await self._exit()
        logger.info('Exited Application')

    async def abort(self) -> None:
        await self._exit(force=True)
        logger.info('Aborted Application')

    async def _exit(self, force: bool = False) -> None:
        await self._task_manager.stop_all_tasks(force=force)
        await self._task_manager.destroy_all_tasks()
        self._destroy()

    async def restart(self) -> None:
        logger.info('Restarting Application...')
        await self.exit()
        await self.launch()

    def has_task(self, room_id: int) -> bool:
        return self._task_manager.has_task(room_id)

    async def add_task(self, room_id: int) -> int:
        room_id = await ensure_room_id(room_id)

        if self._task_manager.has_task(room_id):
            raise ExistsError(
                f'a task for the room {room_id} is already existed'
            )

        settings = self._settings_manager.find_task_settings(room_id)
        if not settings:
            settings = await self._settings_manager.add_task_settings(room_id)

        await self._task_manager.add_task(settings)

        return room_id

    async def remove_task(self, room_id: int) -> None:
        logger.info(f'Removing task {room_id}...')
        await self._task_manager.remove_task(room_id)
        await self._settings_manager.remove_task_settings(room_id)
        logger.info(f'Successfully removed task {room_id}')

    async def remove_all_tasks(self) -> None:
        logger.info('Removing all tasks...')
        await self._task_manager.remove_all_tasks()
        await self._settings_manager.remove_all_task_settings()
        logger.info('Successfully removed all tasks')

    async def start_task(self, room_id: int) -> None:
        logger.info(f'Starting task {room_id}...')
        await self._task_manager.start_task(room_id)
        await self._settings_manager.mark_task_enabled(room_id)
        logger.info(f'Successfully started task {room_id}')

    async def stop_task(self, room_id: int, force: bool = False) -> None:
        logger.info(f'Stopping task {room_id}...')
        await self._task_manager.stop_task(room_id, force)
        await self._settings_manager.mark_task_disabled(room_id)
        logger.info(f'Successfully stopped task {room_id}')

    async def start_all_tasks(self) -> None:
        logger.info('Starting all tasks...')
        await self._task_manager.start_all_tasks()
        await self._settings_manager.mark_all_tasks_enabled()
        logger.info('Successfully started all tasks')

    async def stop_all_tasks(self, force: bool = False) -> None:
        logger.info('Stopping all tasks...')
        await self._task_manager.stop_all_tasks(force)
        await self._settings_manager.mark_all_tasks_disabled()
        logger.info('Successfully stopped all tasks')

    async def enable_task_monitor(self, room_id: int) -> None:
        logger.info(f'Enabling monitor for task {room_id}...')
        await self._task_manager.enable_task_monitor(room_id)
        await self._settings_manager.mark_task_monitor_enabled(room_id)
        logger.info(f'Successfully enabled monitor for task {room_id}')

    async def disable_task_monitor(self, room_id: int) -> None:
        logger.info(f'Disabling monitor for task {room_id}...')
        await self._task_manager.disable_task_monitor(room_id)
        await self._settings_manager.mark_task_monitor_disabled(room_id)
        logger.info(f'Successfully disabled monitor for task {room_id}')

    async def enable_all_task_monitors(self) -> None:
        logger.info('Enabling monitors for all tasks...')
        await self._task_manager.enable_all_task_monitors()
        await self._settings_manager.mark_all_task_monitors_enabled()
        logger.info('Successfully enabled monitors for all tasks')

    async def disable_all_task_monitors(self) -> None:
        logger.info('Disabling monitors for all tasks...')
        await self._task_manager.disable_all_task_monitors()
        await self._settings_manager.mark_all_task_monitors_disabled()
        logger.info('Successfully disabled monitors for all tasks')

    async def enable_task_recorder(self, room_id: int) -> None:
        logger.info(f'Enabling recorder for task {room_id}...')
        await self._task_manager.enable_task_recorder(room_id)
        await self._settings_manager.mark_task_recorder_enabled(room_id)
        logger.info(f'Successfully enabled recorder for task {room_id}')

    async def disable_task_recorder(
        self, room_id: int, force: bool = False
    ) -> None:
        logger.info(f'Disabling recorder for task {room_id}...')
        await self._task_manager.disable_task_recorder(room_id, force)
        await self._settings_manager.mark_task_recorder_disabled(room_id)
        logger.info(f'Successfully disabled recorder for task {room_id}')

    async def enable_all_task_recorders(self) -> None:
        logger.info('Enabling recorders for all tasks...')
        await self._task_manager.enable_all_task_recorders()
        await self._settings_manager.mark_all_task_recorders_enabled()
        logger.info('Successfully enabled recorders for all tasks')

    async def disable_all_task_recorders(self, force: bool = False) -> None:
        logger.info('Disabling recorders for all tasks...')
        await self._task_manager.disable_all_task_recorders(force)
        await self._settings_manager.mark_all_task_recorders_disabled()
        logger.info('Successfully disabled recorders for all tasks')

    def get_task_data(self, room_id: int) -> TaskData:
        return self._task_manager.get_task_data(room_id)

    def get_all_task_data(self) -> Iterator[TaskData]:
        yield from self._task_manager.get_all_task_data()

    def get_task_param(self, room_id: int) -> TaskParam:
        return self._task_manager.get_task_param(room_id)

    def get_task_metadata(self, room_id: int) -> Optional[MetaData]:
        return self._task_manager.get_task_metadata(room_id)

    def get_task_stream_profile(self, room_id: int) -> StreamProfile:
        return self._task_manager.get_task_stream_profile(room_id)

    def get_task_video_file_details(
        self, room_id: int
    ) -> Iterator[VideoFileDetail]:
        yield from self._task_manager.get_task_video_file_details(room_id)

    def get_task_danmaku_file_details(
        self, room_id: int
    ) -> Iterator[DanmakuFileDetail]:
        yield from self._task_manager.get_task_danmaku_file_details(room_id)

    def can_cut_stream(self, room_id: int) -> bool:
        return self._task_manager.can_cut_stream(room_id)

    def cut_stream(self, room_id: int) -> bool:
        return self._task_manager.cut_stream(room_id)

    async def update_task_info(self, room_id: int) -> None:
        logger.info(f'Updating info for task {room_id}...')
        await self._task_manager.update_task_info(room_id)
        logger.info(f'Successfully updated info for task {room_id}')

    async def update_all_task_infos(self) -> None:
        logger.info('Updating info for all tasks...')
        await self._task_manager.update_all_task_infos()
        logger.info('Successfully updated info for all tasks')

    def get_settings(
        self,
        include: Optional[KeySetOfSettings] = None,
        exclude: Optional[KeySetOfSettings] = None,
    ) -> SettingsOut:
        return self._settings_manager.get_settings(include, exclude)

    async def change_settings(self, settings: SettingsIn) -> SettingsOut:
        return await self._settings_manager.change_settings(settings)

    def get_task_options(self, room_id: int) -> TaskOptions:
        return self._settings_manager.get_task_options(room_id)

    async def change_task_options(
        self, room_id: int, options: TaskOptions
    ) -> TaskOptions:
        return await self._settings_manager.change_task_options(
            room_id, options
        )

    def _setup(self) -> None:
        self._setup_logger()
        self._setup_exception_handler()
        self._setup_space_monitor()
        self._setup_space_event_submitter()
        self._setup_space_reclaimer()
        self._setup_notifiers()
        self._setup_webhooks()

    def _setup_logger(self) -> None:
        self._settings_manager.apply_logging_settings()

    def _setup_exception_handler(self) -> None:
        self._exception_handler = ExceptionHandler()
        self._exception_handler.enable()

    def _setup_space_monitor(self) -> None:
        self._space_monitor = SpaceMonitor(self._out_dir)
        self._settings_manager.apply_space_monitor_settings()
        self._space_monitor.enable()

    def _setup_space_event_submitter(self) -> None:
        self._space_event_submitter = SpaceEventSubmitter(self._space_monitor)

    def _setup_space_reclaimer(self) -> None:
        self._space_reclaimer = SpaceReclaimer(
            self._space_monitor, self._out_dir,
        )
        self._settings_manager.apply_space_reclaimer_settings()
        self._space_reclaimer.enable()

    def _setup_notifiers(self) -> None:
        self._email_notifier = EmailNotifier()
        self._serverchan_notifier = ServerchanNotifier()
        self._pushdeer_notifier = PushdeerNotifier()
        self._pushplus_notifier = PushplusNotifier()
        self._telegram_notifier = TelegramNotifier()
        self._settings_manager.apply_email_notification_settings()
        self._settings_manager.apply_serverchan_notification_settings()
        self._settings_manager.apply_pushdeer_notification_settings()
        self._settings_manager.apply_pushplus_notification_settings()
        self._settings_manager.apply_telegram_notification_settings()

    def _setup_webhooks(self) -> None:
        self._webhook_emitter = WebHookEmitter()
        self._settings_manager.apply_webhooks_settings()
        self._webhook_emitter.enable()

    def _destroy(self) -> None:
        self._destroy_space_reclaimer()
        self._destroy_space_event_submitter()
        self._destroy_space_monitor()
        self._destroy_notifiers()
        self._destroy_webhooks()
        self._destroy_exception_handler()

    def _destroy_space_monitor(self) -> None:
        self._space_monitor.disable()
        del self._space_monitor

    def _destroy_space_event_submitter(self) -> None:
        del self._space_event_submitter

    def _destroy_space_reclaimer(self) -> None:
        self._space_reclaimer.disable()
        del self._space_reclaimer

    def _destroy_notifiers(self) -> None:
        self._email_notifier.disable()
        self._serverchan_notifier.disable()
        self._pushdeer_notifier.disable()
        self._pushplus_notifier.disable()
        self._telegram_notifier.disable()
        del self._email_notifier
        del self._serverchan_notifier
        del self._pushdeer_notifier
        del self._pushplus_notifier
        del self._telegram_notifier

    def _destroy_webhooks(self) -> None:
        self._webhook_emitter.disable()
        del self._webhook_emitter

    def _destroy_exception_handler(self) -> None:
        self._exception_handler.disable()
        del self._exception_handler



from .event_center import EventCenter
from .event_emitter import EventEmitter, EventListener
from .models import (
    BaseEvent,
    BaseEventData,
    LiveBeganEvent,
    LiveBeganEventData,
    LiveEndedEvent,
    LiveEndedEventData,
    RoomChangeEvent,
    RoomChangeEventData,
    RecordingStartedEvent,
    RecordingStartedEventData,
    RecordingFinishedEvent,
    RecordingFinishedEventData,
    RecordingCancelledEvent,
    RecordingCancelledEventData,
    VideoFileCreatedEvent,
    VideoFileCreatedEventData,
    VideoFileCompletedEvent,
    VideoFileCompletedEventData,
    DanmakuFileCreatedEvent,
    DanmakuFileCreatedEventData,
    DanmakuFileCompletedEvent,
    DanmakuFileCompletedEventData,
    RawDanmakuFileCreatedEvent,
    RawDanmakuFileCreatedEventData,
    RawDanmakuFileCompletedEvent,
    RawDanmakuFileCompletedEventData,
    VideoPostprocessingCompletedEvent,
    VideoPostprocessingCompletedEventData,
    SpaceNoEnoughEvent,
    SpaceNoEnoughEventData,
    Error,
    ErrorData,
)


__all__ = (
    'EventCenter',
    'EventEmitter', 
    'EventListener',

    'BaseEvent',
    'BaseEventData',
    'LiveBeganEvent',
    'LiveBeganEventData',
    'LiveEndedEvent',
    'LiveEndedEventData',
    'RoomChangeEvent',
    'RoomChangeEventData',
    'RecordingStartedEvent',
    'RecordingStartedEventData',
    'RecordingFinishedEvent',
    'RecordingFinishedEventData',
    'RecordingCancelledEvent',
    'RecordingCancelledEventData',
    'VideoFileCreatedEvent',
    'VideoFileCreatedEventData',
    'VideoFileCompletedEvent',
    'VideoFileCompletedEventData',
    'DanmakuFileCreatedEvent',
    'DanmakuFileCreatedEventData',
    'DanmakuFileCompletedEvent',
    'DanmakuFileCompletedEventData',
    'RawDanmakuFileCreatedEvent',
    'RawDanmakuFileCreatedEventData',
    'RawDanmakuFileCompletedEvent',
    'RawDanmakuFileCompletedEventData',
    'VideoPostprocessingCompletedEvent',
    'VideoPostprocessingCompletedEventData',
    'SpaceNoEnoughEvent',
    'SpaceNoEnoughEventData',
    'Error',
    'ErrorData',
)



from __future__ import annotations

import logging
from typing import Optional

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from tqdm import tqdm

from .typing import FLVStream, FLVStreamItem

__all__ = ('ProgressBar',)


logger = logging.getLogger(__name__)


class ProgressBar:
    def __init__(
        self,
        desc: str,
        postfix: Optional[str] = None,
        total: Optional[int] = None,
        disable: Optional[bool] = False,
    ) -> None:
        self._desc = desc
        self._postfix = postfix
        self._total = total
        self._disable = disable
        self._pbar: Optional[tqdm] = None

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._progress(source)

    def _progress(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            subscription = SerialDisposable()

            self._pbar = tqdm(
                disable=self._disable,
                desc=self._desc,
                total=self._total,
                unit='B',
                unit_scale=True,
                unit_divisor=1024,
                postfix=self._postfix,
            )
            self._pbar.disable

            def on_next(item: FLVStreamItem) -> None:
                if self._pbar is not None:
                    self._pbar.update(len(item))
                observer.on_next(item)

            def on_completed() -> None:
                if self._pbar is not None:
                    self._pbar.close()
                    self._pbar = None
                observer.on_completed()

            def dispose() -> None:
                if self._pbar is not None:
                    self._pbar.close()
                    self._pbar = None

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)



from __future__ import annotations

import io
import logging
from typing import List, Optional, Union

from reactivex import Observable, Subject, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from blrec.utils.ffprobe import StreamProfile, ffprobe

from .segment_fetcher import InitSectionData, SegmentData

__all__ = ('Prober', 'StreamProfile')


logger = logging.getLogger(__name__)


class Prober:
    def __init__(self) -> None:
        self._profiles: Subject[StreamProfile] = Subject()

    def _reset(self) -> None:
        self._gathering: bool = False
        self._gathered_items: List[Union[InitSectionData, SegmentData]] = []

    @property
    def profiles(self) -> Observable[StreamProfile]:
        return self._profiles

    def __call__(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        return self._probe(source)

    def _probe(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        def subscribe(
            observer: abc.ObserverBase[Union[InitSectionData, SegmentData]],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            self._reset()

            def on_next(item: Union[InitSectionData, SegmentData]) -> None:
                if isinstance(item, InitSectionData):
                    self._gathered_items.clear()
                    self._gathering = True

                if self._gathering:
                    self._gathered_items.append(item)
                    if len(self._gathered_items) >= 10:
                        try:
                            self._do_probe()
                        except Exception as e:
                            logger.warning(f'Failed to probe stream: {repr(e)}')
                        finally:
                            self._gathered_items.clear()
                            self._gathering = False

                observer.on_next(item)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                self._reset()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _do_probe(self) -> None:
        bytes_io = io.BytesIO()
        for item in self._gathered_items:
            bytes_io.write(item.payload)

        def on_next(profile: StreamProfile) -> None:
            self._profiles.on_next(profile)

        def on_error(e: Exception) -> None:
            logger.warning(f'Failed to probe stream by ffprobe: {repr(e)}')

        ffprobe(bytes_io.getvalue()).subscribe(on_next, on_error)



from __future__ import annotations

from bitarray import bitarray
from bitarray.util import ba2int


__all__ = 'BitsReader',


class BitsReader:
    def __init__(self, bits: bitarray) -> None:
        self._bits = bits
        self._ptr = 0

    def read_bits_as_int(self, n: int) -> int:
        return ba2int(self.read_bits(n))

    def read_bits(self, n: int) -> bitarray:
        result = self.next_bits(n)
        self._ptr += n
        return result

    def next_bits(self, n: int) -> bitarray:
        assert n >= 0
        if n == 0:
            return bitarray('0')
        return self._bits[self._ptr:self._ptr + n]



import os
import re
import shlex
from subprocess import PIPE, Popen
from typing import Any, Final, List, Optional, Union

import attr
from reactivex import Observable, abc, create
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from reactivex.scheduler.currentthreadscheduler import CurrentThreadScheduler
from tqdm import tqdm

__all__ = 'RemuxingResult', 'remux_video'


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RemuxingProgress:
    count: int
    total: int


_ERROR_PATTERN = re.compile(
    r'\b(error|missing|invalid|corrupt|illegal|overflow|out of range)\b', re.IGNORECASE
)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RemuxingResult:

    return_code: int
    output: str

    def is_done(self) -> bool:
        return self.return_code == 0

    def is_successful(self) -> bool:
        return (
            self.is_done()
            and not self.may_timestamps_incorrect()
            and not self.has_errors()
        )

    def is_warned(self) -> bool:
        return self.is_done() and self.may_timestamps_incorrect()

    def is_failed(self) -> bool:
        return not self.is_done() or self.has_errors()

    def has_errors(self) -> bool:
        return _ERROR_PATTERN.search(self.output) is not None

    def may_timestamps_incorrect(self) -> bool:
        return 'Non-monotonous DTS in output stream' in self.output


def remux_video(
    in_path: str,
    out_path: str,
    metadata_path: Optional[str] = None,
    *,
    show_progress: bool = False,
    remove_filler_data: bool = False,
) -> Observable[Union[RemuxingProgress, RemuxingResult]]:
    SIZE_PATTERN: Final = re.compile(r'size=\s*(?P<number>\d+)(?P<unit>[a-zA-Z]?B)')
    if in_path.endswith('.m3u8'):
        total = 0
        for root, dirs, files in os.walk(os.path.dirname(in_path)):
            for filename in files:
                if not (filename.endswith('.m4s') or filename.endswith('.ts')):
                    continue
                total += os.path.getsize(os.path.join(root, filename))
        postfix = os.path.join(
            os.path.basename(os.path.dirname(in_path)), os.path.basename(in_path)
        )
    else:
        total = os.path.getsize(in_path)
        postfix = os.path.basename(in_path)

    def parse_size(line: str) -> int:
        match = SIZE_PATTERN.search(line)
        assert match is not None
        result = match.groupdict()

        unit = result['unit']
        number = int(result['number'])

        if unit == 'B':
            size = number
        elif unit == 'kB':
            size = 1024 * number
        elif unit == 'MB':
            size = 1024**2 * number
        elif unit == 'GB':
            size = 1024**3 * number
        else:
            raise ValueError(unit)

        return size

    def should_output_line(line: str) -> bool:
        line = line.strip()
        return not (
            line.startswith('frame=')
            or line.startswith('Press [q]')
            or (line.startswith('[hls') and 'Opening' in line and 'for reading' in line)
        )

    def subscribe(
        observer: abc.ObserverBase[Union[RemuxingProgress, RemuxingResult]],
        scheduler: Optional[abc.SchedulerBase] = None,
    ) -> abc.DisposableBase:
        _scheduler = scheduler or CurrentThreadScheduler.singleton()

        disposed = False
        cancelable = SerialDisposable()

        def action(scheduler: abc.SchedulerBase, state: Optional[Any] = None) -> None:
            if disposed:
                return

            with tqdm(
                desc='Remuxing',
                total=total,
                unit='B',
                unit_scale=True,
                unit_divisor=1024,
                postfix=postfix,
                disable=not show_progress,
            ) as pbar:
                cmd = f'ffmpeg -i "{in_path}"'
                if metadata_path is not None:
                    cmd += f' -i "{metadata_path}" -map_metadata 1'
                cmd += ' -codec copy'
                if remove_filler_data:
                    # https://forum.doom9.org/showthread.php?t=152051
                    # ISO_IEC_14496-10_2020(E)
                    # Table 7-1 – NAL unit type codes, syntax element categories, and NAL unit type classes  # noqa
                    # 7.4.2.7 Filler data RBSP semantics
                    cmd += ' -bsf:v filter_units=remove_types=12'
                cmd += f' "{out_path}" -y'

                args = shlex.split(cmd)
                out_lines: List[str] = []

                try:
                    with Popen(
                        args, stderr=PIPE, encoding='utf8', errors='backslashreplace'
                    ) as process:
                        assert process.stderr is not None
                        while not disposed:
                            line = process.stderr.readline()
                            if not line:
                                if process.poll() is not None:
                                    break
                                else:
                                    continue

                            if line.startswith('frame='):
                                size = parse_size(line)
                                pbar.update(size - pbar.n)
                                progress = RemuxingProgress(size, total)
                                observer.on_next(progress)

                            if should_output_line(line):
                                out_lines.append(line)

                        if not disposed and process.returncode == 0:
                            pbar.update(total)
                            progress = RemuxingProgress(total, total)
                            observer.on_next(progress)
                except Exception as e:
                    observer.on_error(e)
                else:
                    result = RemuxingResult(process.returncode, ''.join(out_lines))
                    observer.on_next(result)
                    observer.on_completed()

        cancelable.disposable = _scheduler.schedule(action)

        def dispose() -> None:
            nonlocal disposed
            disposed = True

        return CompositeDisposable(cancelable, Disposable(dispose))

    return create(subscribe)



from __future__ import annotations

import logging
import time
from typing import Optional, TypeVar

import aiohttp
import requests
from reactivex import Observable, abc

from blrec.bili.live import Live
from blrec.utils import operators as utils_ops
from blrec.utils.mixins import AsyncCooperationMixin

__all__ = ('ConnectionErrorHandler',)


logger = logging.getLogger(__name__)

_T = TypeVar('_T')


class ConnectionErrorHandler(AsyncCooperationMixin):
    def __init__(
        self,
        live: Live,
        *,
        disconnection_timeout: Optional[int] = None,
        check_interval: int = 3,
    ) -> None:
        super().__init__()
        self._live = live
        self.disconnection_timeout = disconnection_timeout or 600  # seconds
        self.check_interval = check_interval

    def __call__(self, source: Observable[_T]) -> Observable[_T]:
        return self._handle(source).pipe(
            utils_ops.retry(should_retry=self._should_retry)
        )

    def _handle(self, source: Observable[_T]) -> Observable[_T]:
        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_error(exc: Exception) -> None:
                try:
                    raise exc
                except (
                    aiohttp.ClientConnectionError,
                    requests.exceptions.ConnectionError,
                ) as e:
                    logger.warning(repr(e))
                    if self._wait_for_connection_error():
                        observer.on_error(exc)
                    else:
                        observer.on_completed()
                except Exception:
                    pass
                observer.on_error(exc)

            return source.subscribe(
                observer.on_next, on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _should_retry(self, exc: Exception) -> bool:
        if isinstance(
            exc, (aiohttp.ClientConnectionError, requests.exceptions.ConnectionError)
        ):
            return True
        else:
            return False

    def _wait_for_connection_error(self) -> bool:
        timeout = self.disconnection_timeout
        logger.info(f'Waiting {timeout} seconds for connection recovery... ')
        timebase = time.monotonic()
        while not self._run_coroutine(self._live.check_connectivity()):
            if timeout is not None and time.monotonic() - timebase > timeout:
                logger.error(f'Connection not recovered in {timeout} seconds')
                return False
            time.sleep(self.check_interval)
        else:
            logger.info('Connection recovered')
            return True



from __future__ import annotations

import io
import logging

from reactivex import Observable
from reactivex import operators as ops

from blrec.flv import operators as flv_ops
from blrec.flv.exceptions import FlvDataError
from blrec.flv.operators.typing import FLVStream
from blrec.utils import operators as utils_ops

from ..stream_param_holder import StreamParamHolder

__all__ = ('StreamParser',)


logger = logging.getLogger(__name__)


class StreamParser:
    def __init__(
        self,
        stream_param_holder: StreamParamHolder,
        *,
        ignore_eof: bool = False,
        ignore_value_error: bool = False,
    ) -> None:
        self._stream_param_holder = stream_param_holder
        self._ignore_eof = ignore_eof
        self._ignore_value_error = ignore_value_error

    def __call__(self, source: Observable[io.RawIOBase]) -> FLVStream:
        return source.pipe(  # type: ignore
            flv_ops.parse(
                ignore_eof=self._ignore_eof,
                ignore_value_error=self._ignore_value_error,
                backup_timestamp=True,
            ),
            ops.do_action(on_error=self._before_retry),
            utils_ops.retry(should_retry=self._should_retry),
        )

    def _should_retry(self, exc: Exception) -> bool:
        if isinstance(exc, (EOFError, FlvDataError)):
            return True
        else:
            return False

    def _before_retry(self, exc: Exception) -> None:
        try:
            raise exc
        except EOFError:
            logger.debug(repr(exc))
        except FlvDataError:
            logger.warning(f'Failed to parse stream: {repr(exc)}')
            if not self._stream_param_holder.use_alternative_stream:
                self._stream_param_holder.use_alternative_stream = True
            else:
                self._stream_param_holder.use_alternative_stream = False
                self._stream_param_holder.rotate_api_platform()
        except Exception:
            pass



import time


__all__ = 'Statistics',


class Statistics:
    def __init__(self, interval: float = 1.0) -> None:
        self._interval = interval
        self._frozen = True

        self._count: int = 0
        self._rate: float = 0.0
        self._start_time: float = 0.0
        self._last_time: float = 0.0
        self._last_count: int = 0
        self._elapsed: float = 0.0

    @property
    def count(self) -> int:
        return self._count

    @property
    def rate(self) -> float:
        if self._frozen:
            return self._rate

        curr_time = time.monotonic()
        time_delta = curr_time - self._last_time

        if time_delta >= self._interval:
            count_delta = self._count - self._last_count
            self._rate = count_delta / time_delta

            self._last_time = curr_time
            self._last_count = self._count

        return self._rate

    @property
    def elapsed(self) -> float:
        if not self._frozen:
            self._elapsed = time.monotonic() - self._start_time
        return self._elapsed

    def freeze(self) -> None:
        self._frozen = True

    def reset(self) -> None:
        self._count = self._last_count = 0
        self._start_time = self._last_time = time.monotonic()
        self._rate = 0.0
        self._elapsed = 0.0

        self._frozen = False

    def submit(self, count: int) -> None:
        if not self._frozen:
            self._count += count



import os
import logging
from logging import LogRecord, Handler
from logging.handlers import RotatingFileHandler
from datetime import datetime
import asyncio
import threading
import atexit
from typing import Any, List, Optional

from colorama import init, deinit, Fore, Back, Style
from tqdm import tqdm

from .typing import LOG_LEVEL


__all__ = 'configure_logger', 'ConsoleHandler', 'TqdmOutputStream'


class TqdmOutputStream:
    def write(self, string: str = '') -> None:
        tqdm.write(string, end='')


class ConsoleHandler(logging.StreamHandler):
    def __init__(self, stream=None) -> None:  # type: ignore
        super().__init__(stream)

    def format(self, record: LogRecord) -> str:
        msg = super().format(record)

        level = record.levelno
        if level == logging.DEBUG:
            style = Fore.GREEN
        elif level == logging.WARNING:
            style = Fore.YELLOW
        elif level == logging.ERROR:
            style = Fore.RED
        elif level == logging.CRITICAL:
            style = Fore.WHITE + Back.RED + Style.BRIGHT
        else:
            style = ''

        return style + msg + Style.RESET_ALL if style else msg


_old_factory = logging.getLogRecordFactory()


def obtain_room_id() -> str:
    try:
        task = asyncio.current_task()
        assert task is not None
    except Exception:
        name = threading.current_thread().getName()
    else:
        name = task.get_name()

    if '::' in name:
        if (room_id := name.split('::')[-1]):
            return room_id

    return ''


def record_factory(*args: Any, **kwargs: Any) -> LogRecord:
    record = _old_factory(*args, **kwargs)

    if (room_id := obtain_room_id()):
        record.roomid = '[' + room_id + '] '  # type: ignore
    else:
        record.roomid = ''  # type: ignore

    return record


logging.setLogRecordFactory(record_factory)


_old_handlers: List[Handler] = []


def configure_logger(
    log_dir: str,
    *,
    console_log_level: LOG_LEVEL = 'INFO',
    max_bytes: Optional[int] = None,
    backup_count: Optional[int] = None,
) -> None:
    # config root logger
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # config formatter
    formatter = logging.Formatter(
        '[%(asctime)s] [%(levelname)s] [%(module)s] %(roomid)s%(message)s'
    )

    # logging to console
    console_handler = ConsoleHandler(TqdmOutputStream())
    console_handler.setLevel(logging.getLevelName(console_log_level))
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    # logging to file
    log_file_path = make_log_file_path(log_dir)
    file_handler = RotatingFileHandler(
        log_file_path,
        maxBytes=max_bytes or 1024 ** 2 * 10,
        backupCount=backup_count or 1,
        encoding='utf-8',
    )
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    # remove old handlers after re-configured
    for handler in _old_handlers:
        logger.removeHandler(handler)

    # retain old handlers for the removing
    _old_handlers.append(console_handler)
    _old_handlers.append(file_handler)

    logger.info(f'log file: {log_file_path}')


def make_log_file_path(log_dir: str) -> str:
    data_time_string = datetime.now().strftime("%Y-%m-%d-%H%M%S-%f")
    filename = f'blrec_{data_time_string}.log'
    return os.path.abspath(os.path.join(log_dir, filename))


init()
atexit.register(deinit)



from .analyse import Analyser, KeyFrames, MetaData
from .concat import JoinPoint, JoinPointData, JoinPointExtractor, concat
from .correct import correct
from .cut import Cutter
from .defragment import defragment
from .dump import Dumper
from .fix import fix
from .inject import Injector
from .limit import Limiter
from .parse import parse
from .probe import Prober, StreamProfile
from .process import process
from .progress import ProgressBar
from .sort import sort
from .split import split

__all__ = (
    'Analyser',
    'concat',
    'concat',
    'correct',
    'Cutter',
    'defragment',
    'Dumper',
    'fix',
    'Injector',
    'JoinPoint',
    'JoinPointData',
    'JoinPointExtractor',
    'KeyFrames',
    'Limiter',
    'MetaData',
    'parse',
    'Prober',
    'process',
    'ProgressBar',
    'sort',
    'split',
    'StreamProfile',
)



import asyncio
import logging
from enum import Enum
from threading import Lock
from typing import Set

import aiofiles
import aiohttp
from tenacity import retry, stop_after_attempt, wait_fixed

from blrec.bili.live import Live
from blrec.exception import exception_callback
from blrec.logging.room_id import aio_task_with_room_id
from blrec.path import cover_path
from blrec.utils.hash import sha1sum
from blrec.utils.mixins import SwitchableMixin

from .stream_recorder import StreamRecorder, StreamRecorderEventListener

__all__ = ('CoverDownloader',)


logger = logging.getLogger(__name__)


class CoverSaveStrategy(Enum):
    DEFAULT = 'default'
    DEDUP = 'dedup'

    def __str__(self) -> str:
        return self.value

    # workaround for value serialization
    def __repr__(self) -> str:
        return str(self)


class CoverDownloader(StreamRecorderEventListener, SwitchableMixin):
    def __init__(
        self,
        live: Live,
        stream_recorder: StreamRecorder,
        *,
        save_cover: bool = False,
        cover_save_strategy: CoverSaveStrategy = CoverSaveStrategy.DEFAULT,
    ) -> None:
        super().__init__()
        self._live = live
        self._stream_recorder = stream_recorder
        self._lock: Lock = Lock()
        self._sha1_set: Set[str] = set()
        self.save_cover = save_cover
        self.cover_save_strategy = cover_save_strategy

    def _do_enable(self) -> None:
        self._sha1_set.clear()
        self._stream_recorder.add_listener(self)
        logger.debug('Enabled cover downloader')

    def _do_disable(self) -> None:
        self._stream_recorder.remove_listener(self)
        logger.debug('Disabled cover downloader')

    async def on_video_file_completed(self, video_path: str) -> None:
        with self._lock:
            if not self.save_cover:
                return
            task = asyncio.create_task(self._save_cover(video_path))
            task.add_done_callback(exception_callback)

    @aio_task_with_room_id
    async def _save_cover(self, video_path: str) -> None:
        try:
            await self._live.update_room_info()
            cover_url = self._live.room_info.cover
            data = await self._fetch_cover(cover_url)
            sha1 = sha1sum(data)
            if (
                self.cover_save_strategy == CoverSaveStrategy.DEDUP
                and sha1 in self._sha1_set
            ):
                return
            path = cover_path(video_path, ext=cover_url.rsplit('.', 1)[-1])
            await self._save_file(path, data)
            self._sha1_set.add(sha1)
        except Exception as e:
            logger.error(f'Failed to save cover image: {repr(e)}')
        else:
            logger.info(f'Saved cover image: {path}')

    @retry(reraise=True, wait=wait_fixed(1), stop=stop_after_attempt(3))
    async def _fetch_cover(self, url: str) -> bytes:
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(url) as response:
                return await response.read()

    async def _save_file(self, path: str, data: bytes) -> None:
        async with aiofiles.open(path, 'wb') as file:
            await file.write(data)



import logging
from typing import Optional

from reactivex import operators as ops
from reactivex.scheduler import NewThreadScheduler

from blrec.bili.live import Live
from blrec.bili.typing import QualityNumber
from blrec.hls import operators as hls_ops
from blrec.hls.metadata_dumper import MetadataDumper

from . import operators as core_ops
from .stream_recorder_impl import StreamRecorderImpl

__all__ = ('HLSRawStreamRecorderImpl',)


logger = logging.getLogger(__name__)


class HLSRawStreamRecorderImpl(StreamRecorderImpl):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        *,
        quality_number: QualityNumber = 10000,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        disconnection_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__(
            live=live,
            out_dir=out_dir,
            path_template=path_template,
            stream_format='fmp4',
            recording_mode='raw',
            quality_number=quality_number,
            buffer_size=buffer_size,
            read_timeout=read_timeout,
            disconnection_timeout=disconnection_timeout,
            filesize_limit=filesize_limit,
            duration_limit=duration_limit,
        )

        self._playlist_fetcher = hls_ops.PlaylistFetcher(self._live, self._session)
        self._playlist_dumper = hls_ops.PlaylistDumper(self._path_provider)
        self._segment_fetcher = hls_ops.SegmentFetcher(self._live, self._session)
        self._segment_dumper = hls_ops.SegmentDumper(self._playlist_dumper)
        self._ff_metadata_dumper = MetadataDumper(
            self._playlist_dumper, self._metadata_provider
        )

        self._prober = hls_ops.Prober()
        self._dl_statistics = core_ops.SizedStatistics()

        self._recording_monitor = core_ops.RecordingMonitor(
            live, lambda: self._playlist_dumper.duration
        )

        self._prober.profiles.subscribe(self._on_profile_updated)
        self._playlist_dumper.file_opened.subscribe(self._on_video_file_opened)
        self._playlist_dumper.file_closed.subscribe(self._on_video_file_closed)
        self._recording_monitor.interrupted.subscribe(self._on_recording_interrupted)
        self._recording_monitor.recovered.subscribe(self._on_recording_recovered)

    @property
    def recording_path(self) -> Optional[str]:
        return self._playlist_dumper.path

    def _on_start(self) -> None:
        self._ff_metadata_dumper.enable()

    def _on_stop(self) -> None:
        self._ff_metadata_dumper.disable()

    def _run(self) -> None:
        self._subscription = (
            self._stream_param_holder.get_stream_params()  # type: ignore
            .pipe(
                self._stream_url_resolver,
                ops.subscribe_on(
                    NewThreadScheduler(self._thread_factory('PlaylistDownloader'))
                ),
                self._playlist_fetcher,
                self._recording_monitor,
                self._connection_error_handler,
                self._request_exception_handler,
                self._playlist_dumper,
                ops.observe_on(
                    NewThreadScheduler(self._thread_factory('SegmentDownloader'))
                ),
                self._segment_fetcher,
                self._dl_statistics,
                self._prober,
                self._segment_dumper,
                self._rec_statistics,
                self._progress_bar,
                self._exception_handler,
            )
            .subscribe(on_completed=self._on_completed)
        )



from http import HTTPStatus
from typing import Any, Dict, Union


from .schemas import ResponseMessage


ResponseDescriptor = Dict[Union[int, str], Dict[str, Any]]

not_found_responses: ResponseDescriptor = {
    HTTPStatus.NOT_FOUND.value: {
        'description': HTTPStatus.NOT_FOUND.phrase,
        'model': ResponseMessage,
    },
}

forbidden_responses: ResponseDescriptor = {
    HTTPStatus.FORBIDDEN.value: {
        'description': HTTPStatus.FORBIDDEN.phrase,
        'model': ResponseMessage,
    },
}

confict_responses: ResponseDescriptor = {
    HTTPStatus.CONFLICT.value: {
        'description': HTTPStatus.CONFLICT.phrase,
        'model': ResponseMessage,
    },
}

created_responses: ResponseDescriptor = {
    HTTPStatus.CREATED.value: {
        'description': HTTPStatus.CREATED.phrase,
        'model': ResponseMessage,
    },
}

accepted_responses: ResponseDescriptor = {
    HTTPStatus.ACCEPTED.value: {
        'description': HTTPStatus.ACCEPTED.phrase,
        'model': ResponseMessage,
    },
}



import io
import logging

from reactivex import of

from .parse import parse
from .typing import FLVStream

__all__ = ('from_file', 'from_stream')

logger = logging.getLogger(__name__)


def from_stream(
    stream: io.RawIOBase,
    *,
    complete_on_eof: bool = False,
    backup_timestamp: bool = False,
    restore_timestamp: bool = False,
) -> FLVStream:
    return of(stream).pipe(
        parse(
            complete_on_eof=complete_on_eof,
            backup_timestamp=backup_timestamp,
            restore_timestamp=restore_timestamp,
        )
    )


def from_file(
    path: str, *, backup_timestamp: bool = False, restore_timestamp: bool = False
) -> FLVStream:
    return from_stream(
        open(path, 'rb'),
        complete_on_eof=True,
        backup_timestamp=backup_timestamp,
        restore_timestamp=restore_timestamp,
    )



import threading
import asyncio
from functools import wraps
from typing import Any, Awaitable, Callable, TypeVar


_T = TypeVar('_T')


def with_room_id(room_id: int):  # type: ignore
    def decorate(func: Callable[..., _T]) -> Callable[..., _T]:
        @wraps(func)
        def wrapper(*arg: Any, **kwargs: Any) -> _T:
            curr_thread = threading.current_thread()
            old_name = curr_thread.getName()
            curr_thread.setName(f'{func.__qualname__}::{room_id}')
            try:
                return func(*arg, **kwargs)
            finally:
                curr_thread.setName(old_name)
        return wrapper
    return decorate


def aio_task_with_room_id(
    func: Callable[..., Awaitable[_T]]
) -> Callable[..., Awaitable[_T]]:
    @wraps(func)
    async def wrapper(obj: Any, *arg: Any, **kwargs: Any) -> _T:
        if hasattr(obj, '_room_id'):
            room_id = obj._room_id
        elif hasattr(obj, '_live'):
            room_id = obj._live.room_id
        else:
            room_id = ''

        curr_task = asyncio.current_task()
        assert curr_task is not None
        old_name = curr_task.get_name()
        curr_task.set_name(f'{func.__qualname__}::{room_id}')
        try:
            return await func(obj, *arg, **kwargs)
        finally:
            curr_task.set_name(old_name)

    return wrapper



from __future__ import annotations

import logging
import os
from typing import Optional

import m3u8
import urllib3
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

__all__ = ('PlaylistResolver',)


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class PlaylistResolver:
    def __call__(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        return self._solve(source)

    def _name_of(self, uri: str) -> str:
        name, ext = os.path.splitext(uri)
        return name

    def _sequence_number_of(self, uri: str) -> int:
        return int(self._name_of(uri))

    def _solve(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        def subscribe(
            observer: abc.ObserverBase[m3u8.Segment],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            last_sequence_number: Optional[int] = None

            def on_next(playlist: m3u8.M3U8) -> None:
                nonlocal last_sequence_number

                if playlist.is_endlist:
                    logger.debug('Playlist ended')

                for seg in playlist.segments:
                    uri = seg.uri
                    name = self._name_of(uri)
                    num = int(name)
                    if last_sequence_number is not None:
                        if last_sequence_number >= num:
                            continue
                        if last_sequence_number + 1 != num:
                            logger.warning(
                                'Segments discontinuous: '
                                f'last sequence number: {last_sequence_number}, '
                                f'current sequence number: {num}'
                            )
                            seg.discontinuity = True
                    observer.on_next(seg)
                    last_sequence_number = num

            def dispose() -> None:
                nonlocal disposed
                nonlocal last_sequence_number
                disposed = True
                last_sequence_number = None

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)



from .space_monitor import SpaceMonitor, SpaceEventListener
from .space_reclaimer import SpaceReclaimer
from .models import DiskUsage
from .helpers import is_space_enough, delete_file


__all__ = (
    'SpaceMonitor',
    'SpaceEventListener',
    'SpaceReclaimer',
    'DiskUsage',

    'is_space_enough',
    'delete_file',
)



import os
import logging
from copy import deepcopy
from typing import Optional

import uvicorn
from uvicorn.config import LOGGING_CONFIG
import typer

from .. import __prog__, __version__
from ..logging import TqdmOutputStream


logger = logging.getLogger(__name__)

cli = typer.Typer()


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f'Bilibili live streaming recorder {__version__}')
        raise typer.Exit()


@cli.command()
def cli_main(
    version: Optional[bool] = typer.Option(
        None,
        '--version',
        callback=version_callback,
        is_eager=True,
        help=f"show {__prog__}'s version and exit",
    ),
    config: str = typer.Option(
        None,
        '--config',
        '-c',
        help='path of settings.toml file',
    ),
    out_dir: Optional[str] = typer.Option(
        None,
        '--out-dir',
        '-o',
        help='path of directory to store record files (overwrite setting)'
    ),
    log_dir: Optional[str] = typer.Option(
        None,
        '--log-dir',
        help='path of directory to store log files (overwrite setting)'
    ),
    host: str = typer.Option('localhost', help='webapp host bind'),
    port: int = typer.Option(2233, help='webapp port bind'),
    open: bool = typer.Option(False, help='open webapp in default browser'),
    key_file: Optional[str] = typer.Option(None, help='SSL key file'),
    cert_file: Optional[str] = typer.Option(None, help='SSL certificate file'),
    api_key: Optional[str] = typer.Option(None, help='web api key'),
) -> None:
    """Bilibili live streaming recorder"""
    if config is not None:
        os.environ['config'] = config
    if api_key is not None:
        os.environ['api_key'] = api_key
    if out_dir is not None:
        os.environ['out_dir'] = out_dir
    if log_dir is not None:
        os.environ['log_dir'] = log_dir

    if open:
        typer.launch(f'http://localhost:{port}')

    logging_config = deepcopy(LOGGING_CONFIG)
    logging_config['handlers']['default']['stream'] = TqdmOutputStream
    logging_config['handlers']['access']['stream'] = TqdmOutputStream

    uvicorn.run(
        'blrec.web:app',
        host=host,
        port=port,
        ssl_keyfile=key_file,
        ssl_certfile=cert_file,
        log_config=logging_config,
        log_level='info',
        access_log=False,
    )


def main() -> int:
    try:
        cli()
    except KeyboardInterrupt:
        return 1
    except SystemExit:
        return 1
    except BaseException as e:
        logger.exception(e)
        return 2
    else:
        return 0
    finally:
        logger.info('Exit')


if __name__ == '__main__':
    main()



from __future__ import annotations

import logging
from typing import Optional

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ..common import (
    is_audio_sequence_header,
    is_metadata_tag,
    is_video_nalu_keyframe,
    is_video_sequence_header,
)
from ..models import BACK_POINTER_SIZE, AudioTag, FlvHeader, FlvTag, ScriptTag, VideoTag
from .correct import correct
from .typing import FLVStream, FLVStreamItem

__all__ = ('Limiter',)


logger = logging.getLogger(__name__)


class Limiter:
    def __init__(
        self,
        filesize_limit: int = 0,  # file size in bytes, no limit by default.
        duration_limit: int = 0,  # duration in seconds, no limit by default.
    ) -> None:
        self.filesize_limit = filesize_limit
        self.duration_limit = duration_limit
        self._reset()

    def _reset(self) -> None:
        self._filesize: int = 0
        self._duration: float = 0.0
        self._max_size_between_keyframes: int = 0
        self._max_duration_between_keyframes: float = 0.0
        self._first_keyframe_tag: Optional[VideoTag] = None
        self._last_keyframe_tag: Optional[VideoTag] = None
        self._last_flv_header: Optional[FlvHeader] = None
        self._last_metadata_tag: Optional[ScriptTag] = None
        self._last_audio_sequence_header: Optional[AudioTag] = None
        self._last_video_sequence_header: Optional[VideoTag] = None

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._limit(source).pipe(correct())

    def _limit(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            def on_next(item: FLVStreamItem) -> None:
                if isinstance(item, FlvHeader):
                    self._reset()
                    self._update_flv_header(item)
                else:
                    self._update_meta_tags(item)
                    will_over_limts = self._check_limits(item)
                    if will_over_limts:
                        self._insert_header_and_tags(observer)
                observer.on_next(item)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                self._reset()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    def _insert_header_and_tags(
        self, observer: abc.ObserverBase[FLVStreamItem]
    ) -> None:
        assert self._last_flv_header is not None
        observer.on_next(self._last_flv_header)
        if self._last_metadata_tag is not None:
            observer.on_next(self._last_metadata_tag)
        if self._last_video_sequence_header is not None:
            observer.on_next(self._last_video_sequence_header)
        if self._last_audio_sequence_header is not None:
            observer.on_next(self._last_audio_sequence_header)

        self._filesize = self._last_flv_header.size
        if self._last_metadata_tag is not None:
            self._filesize += self._last_metadata_tag.tag_size
        if self._last_video_sequence_header is not None:
            self._filesize += self._last_video_sequence_header.tag_size
        if self._last_audio_sequence_header is not None:
            self._filesize += self._last_audio_sequence_header.tag_size
        self._duration = 0.0
        self._first_keyframe_tag = self._last_keyframe_tag

    def _will_filesize_over_limit(self) -> bool:
        return self._filesize + self._max_size_between_keyframes >= self.filesize_limit

    def _will_duration_over_limit(self) -> bool:
        return (
            self._duration + self._max_duration_between_keyframes >= self.duration_limit
        )

    def _update_flv_header(self, header: FlvHeader) -> None:
        self._filesize += header.size + BACK_POINTER_SIZE
        self._last_flv_header = header

    def _update_meta_tags(self, tag: FlvTag) -> None:
        if is_metadata_tag(tag):
            self._last_metadata_tag = tag
        elif is_audio_sequence_header(tag):
            self._last_audio_sequence_header = tag
        elif is_video_sequence_header(tag):
            self._last_video_sequence_header = tag

    def _check_limits(self, tag: FlvTag) -> bool:
        self._filesize += tag.tag_size + BACK_POINTER_SIZE

        if not is_video_nalu_keyframe(tag):
            return False

        if self._first_keyframe_tag is None:
            self._first_keyframe_tag = tag

        if self._last_keyframe_tag is not None:
            self._max_size_between_keyframes = max(
                self._max_size_between_keyframes,
                tag.offset - self._last_keyframe_tag.offset,
            )
            self._max_duration_between_keyframes = max(
                self._max_duration_between_keyframes,
                (tag.timestamp - self._last_keyframe_tag.timestamp) / 1000,
            )

        self._last_keyframe_tag = tag
        self._duration = (
            self._last_keyframe_tag.timestamp - self._first_keyframe_tag.timestamp
        ) / 1000

        if self.filesize_limit > 0 and self._will_filesize_over_limit():
            logger.debug(
                'File size will be over the limit: {} + {}'.format(
                    self._filesize, self._max_size_between_keyframes
                )
            )
            return True

        if self.duration_limit > 0 and self._will_duration_over_limit():
            logger.debug(
                'Duration will be over the limit: {} + {}'.format(
                    self._duration, self._max_duration_between_keyframes
                )
            )
            return True

        return False



import logging
from asyncio import Queue, QueueFull
from typing import Final

from blrec.bili.danmaku_client import DanmakuClient, DanmakuListener
from blrec.bili.typing import Danmaku
from blrec.utils.mixins import StoppableMixin

__all__ = ('RawDanmakuReceiver',)


logger = logging.getLogger(__name__)


class RawDanmakuReceiver(DanmakuListener, StoppableMixin):
    _MAX_QUEUE_SIZE: Final[int] = 2000

    def __init__(self, danmaku_client: DanmakuClient) -> None:
        super().__init__()
        self._danmaku_client = danmaku_client
        self._queue: Queue[Danmaku] = Queue(maxsize=self._MAX_QUEUE_SIZE)

    def _do_start(self) -> None:
        self._danmaku_client.add_listener(self)
        logger.debug('Started raw danmaku receiver')

    def _do_stop(self) -> None:
        self._danmaku_client.remove_listener(self)
        self._clear_queue()
        logger.debug('Stopped raw danmaku receiver')

    async def get_raw_danmaku(self) -> Danmaku:
        return await self._queue.get()

    async def on_danmaku_received(self, danmu: Danmaku) -> None:
        try:
            self._queue.put_nowait(danmu)
        except QueueFull:
            self._queue.get_nowait()  # discard the first item
            self._queue.put_nowait(danmu)

    def _clear_queue(self) -> None:
        self._queue = Queue(maxsize=self._MAX_QUEUE_SIZE)



from typing import Any, Dict, Literal, Mapping


ApiPlatform = Literal[
    'web',
    'android',
]

QualityNumber = Literal[
    20000,  # 4K
    10000,  # 原画
    401,    # 蓝光(杜比)
    400,    # 蓝光
    250,    # 超清
    150,    # 高清
    80,     # 流畅
]

StreamFormat = Literal[
    'flv',
    'ts',
    'fmp4',
]

StreamCodec = Literal[
    'avc',
    'hevc',
]

JsonResponse = Dict[str, Any]
ResponseData = Dict[str, Any]

Danmaku = Mapping[str, Any]



import asyncio
import logging
import time
from typing import Iterator, Optional

from blrec.bili.live import Live
from blrec.bili.typing import QualityNumber, StreamFormat
from blrec.event.event_emitter import EventEmitter
from blrec.flv.operators import MetaData, StreamProfile
from blrec.setting.typing import RecordingMode
from blrec.utils.mixins import AsyncStoppableMixin

from .flv_stream_recorder_impl import FLVStreamRecorderImpl
from .hls_raw_stream_recorder_impl import HLSRawStreamRecorderImpl
from .hls_stream_recorder_impl import HLSStreamRecorderImpl
from .stream_recorder_impl import StreamRecorderEventListener

__all__ = 'StreamRecorder', 'StreamRecorderEventListener'


logger = logging.getLogger(__name__)


class StreamRecorder(
    StreamRecorderEventListener,
    EventEmitter[StreamRecorderEventListener],
    AsyncStoppableMixin,
):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        *,
        stream_format: StreamFormat = 'flv',
        recording_mode: RecordingMode = 'standard',
        quality_number: QualityNumber = 10000,
        fmp4_stream_timeout: int = 10,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        disconnection_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__()

        self._live = live
        self.stream_format = stream_format
        self.recording_mode = recording_mode
        self.fmp4_stream_timeout = fmp4_stream_timeout

        if stream_format == 'flv':
            cls = FLVStreamRecorderImpl
        elif stream_format == 'fmp4':
            if recording_mode == 'standard':
                cls = HLSStreamRecorderImpl  # type: ignore
            else:
                cls = HLSRawStreamRecorderImpl  # type: ignore
        else:
            logger.warning(
                f'The specified stream format ({stream_format}) is '
                'unsupported, will using the stream format (flv) instead.'
            )
            self.stream_format = 'flv'
            cls = FLVStreamRecorderImpl

        self._impl = cls(
            live=live,
            out_dir=out_dir,
            path_template=path_template,
            quality_number=quality_number,
            buffer_size=buffer_size,
            read_timeout=read_timeout,
            disconnection_timeout=disconnection_timeout,
            filesize_limit=filesize_limit,
            duration_limit=duration_limit,
        )

        self._impl.add_listener(self)

    @property
    def stream_url(self) -> str:
        return self._impl.stream_url

    @property
    def stream_host(self) -> str:
        return self._impl.stream_host

    @property
    def record_start_time(self) -> Optional[int]:
        return self._impl.record_start_time

    @property
    def stream_available_time(self) -> Optional[int]:
        return self._impl.stream_available_time

    @stream_available_time.setter
    def stream_available_time(self, ts: Optional[int]) -> None:
        self._impl.stream_available_time = ts

    @property
    def hls_stream_available_time(self) -> Optional[int]:
        return self._impl.hls_stream_available_time

    @hls_stream_available_time.setter
    def hls_stream_available_time(self, ts: Optional[int]) -> None:
        self._impl.hls_stream_available_time = ts

    @property
    def dl_total(self) -> int:
        return self._impl.dl_total

    @property
    def dl_rate(self) -> float:
        return self._impl.dl_rate

    @property
    def rec_elapsed(self) -> float:
        return self._impl.rec_elapsed

    @property
    def rec_total(self) -> int:
        return self._impl.rec_total

    @property
    def rec_rate(self) -> float:
        return self._impl.rec_rate

    @property
    def out_dir(self) -> str:
        return self._impl.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._impl.out_dir = value

    @property
    def path_template(self) -> str:
        return self._impl.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._impl.path_template = value

    @property
    def quality_number(self) -> QualityNumber:
        return self._impl.quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._impl.quality_number = value

    @property
    def real_stream_format(self) -> Optional[StreamFormat]:
        if self.stopped:
            return None
        return self._impl.stream_format

    @property
    def real_quality_number(self) -> Optional[QualityNumber]:
        return self._impl.real_quality_number

    @property
    def buffer_size(self) -> int:
        return self._impl.buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._impl.buffer_size = value

    @property
    def read_timeout(self) -> int:
        return self._impl.read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._impl.read_timeout = value

    @property
    def disconnection_timeout(self) -> int:
        return self._impl.disconnection_timeout

    @disconnection_timeout.setter
    def disconnection_timeout(self, value: int) -> None:
        self._impl.disconnection_timeout = value

    @property
    def filesize_limit(self) -> int:
        return self._impl.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._impl.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._impl.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._impl.duration_limit = value

    @property
    def recording_path(self) -> Optional[str]:
        return self._impl.recording_path

    @property
    def metadata(self) -> Optional[MetaData]:
        return self._impl.metadata

    @property
    def stream_profile(self) -> StreamProfile:
        return self._impl.stream_profile

    def has_file(self) -> bool:
        return self._impl.has_file()

    def get_files(self) -> Iterator[str]:
        yield from self._impl.get_files()

    def clear_files(self) -> None:
        self._impl.clear_files()

    def can_cut_stream(self) -> bool:
        return self._impl.can_cut_stream()

    def cut_stream(self) -> bool:
        return self._impl.cut_stream()

    def update_progress_bar_info(self) -> None:
        self._impl.update_progress_bar_info()

    @property
    def stopped(self) -> bool:
        return self._impl.stopped

    async def _do_start(self) -> None:
        self.hls_stream_available_time = None
        stream_format = self.stream_format
        if stream_format == 'fmp4':
            logger.info('Waiting for the fmp4 stream becomes available...')
            available = await self._wait_fmp4_stream()
            if available:
                if self.stream_available_time is not None:
                    self.hls_stream_available_time = await self._live.get_timestamp()
            else:
                logger.warning(
                    'The specified stream format (fmp4) is not available '
                    f'in {self.fmp4_stream_timeout} seconcds, '
                    'falling back to stream format (flv).'
                )
                stream_format = 'flv'
        self._change_impl(stream_format)
        await self._impl.start()

    async def _do_stop(self) -> None:
        await self._impl.stop()

    async def on_video_file_created(self, path: str, record_start_time: int) -> None:
        await self._emit('video_file_created', path, record_start_time)

    async def on_video_file_completed(self, path: str) -> None:
        await self._emit('video_file_completed', path)

    async def on_stream_recording_interrupted(self, duration: float) -> None:
        await self._emit('stream_recording_interrupted', duration)

    async def on_stream_recording_recovered(self, timestamp: int) -> None:
        await self._emit('stream_recording_recovered', timestamp)

    async def on_stream_recording_completed(self) -> None:
        await self._emit('stream_recording_completed')

    async def _wait_fmp4_stream(self) -> bool:
        end_time = time.monotonic() + self.fmp4_stream_timeout
        available = False  # debounce
        while True:
            try:
                await self._impl._live.get_live_stream_url(stream_format='fmp4')
            except Exception:
                available = False
                if time.monotonic() > end_time:
                    return False
            else:
                if available:
                    return True
                else:
                    available = True
            await asyncio.sleep(1)

    def _change_impl(self, stream_format: StreamFormat) -> None:
        if stream_format == 'flv':
            cls = FLVStreamRecorderImpl
        elif stream_format == 'fmp4':
            if self.recording_mode == 'standard':
                cls = HLSStreamRecorderImpl  # type: ignore
            else:
                cls = HLSRawStreamRecorderImpl  # type: ignore
        else:
            logger.warning(
                f'The specified stream format ({stream_format}) is '
                'unsupported, will using the stream format (flv) instead.'
            )
            cls = FLVStreamRecorderImpl

        if self._impl.__class__ == cls:
            return

        self._impl.remove_listener(self)
        stream_available_time = self._impl.stream_available_time
        hls_stream_available_time = self._impl.hls_stream_available_time

        self._impl = cls(
            live=self._impl._live,
            out_dir=self._impl.out_dir,
            path_template=self._impl.path_template,
            quality_number=self._impl.quality_number,
            buffer_size=self._impl.buffer_size,
            read_timeout=self._impl.read_timeout,
            disconnection_timeout=self._impl.disconnection_timeout,
            filesize_limit=self._impl.filesize_limit,
            duration_limit=self._impl.duration_limit,
        )

        self._impl.add_listener(self)
        self._impl.stream_available_time = stream_available_time
        self._impl.hls_stream_available_time = hls_stream_available_time

        logger.debug(f'Changed stream recorder impl to {cls.__name__}')



from __future__ import annotations

import errno
import logging
from typing import Optional, TypeVar

from reactivex import Observable, abc

from blrec.bili.exceptions import LiveRoomEncrypted, LiveRoomHidden, LiveRoomLocked
from blrec.utils import operators as utils_ops
from blrec.utils.mixins import AsyncCooperationMixin

__all__ = ('ExceptionHandler',)


logger = logging.getLogger(__name__)

_T = TypeVar('_T')


class ExceptionHandler(AsyncCooperationMixin):
    def __call__(self, source: Observable[_T]) -> Observable[_T]:
        return self._handle(source).pipe(utils_ops.retry(delay=1))

    def _handle(self, source: Observable[_T]) -> Observable[_T]:
        def subscribe(
            observer: abc.ObserverBase[_T],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_error(exc: Exception) -> None:
                logger.exception(repr(exc))
                self._submit_exception(exc)
                try:
                    raise exc
                except OSError as e:
                    logger.critical(repr(e), exc_info=e)
                    if e.errno == errno.ENOSPC:
                        # OSError(28, 'No space left on device')
                        observer.on_completed()
                    else:
                        observer.on_error(exc)
                except LiveRoomHidden:
                    logger.error('The live room has been hidden!')
                    observer.on_completed()
                except LiveRoomLocked:
                    logger.error('The live room has been locked!')
                    observer.on_completed()
                except LiveRoomEncrypted:
                    logger.error('The live room has been encrypted!')
                    observer.on_completed()
                except Exception:
                    observer.on_error(exc)

            return source.subscribe(
                observer.on_next, on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)



from __future__ import annotations

from typing import Optional, Sized

from reactivex import Observable, abc

from ..statistics import Statistics

__all__ = ('SizedStatistics',)


class SizedStatistics:
    def __init__(self) -> None:
        self._statistics = Statistics()

    @property
    def count(self) -> int:
        return self._statistics.count

    @property
    def rate(self) -> float:
        return self._statistics.rate

    @property
    def elapsed(self) -> float:
        return self._statistics.elapsed

    def freeze(self) -> None:
        self._statistics.freeze()

    def reset(self) -> None:
        self._statistics.reset()

    def __call__(self, source: Observable[Sized]) -> Observable[Sized]:
        self.reset()
        return self._calc(source)

    def _calc(self, source: Observable[Sized]) -> Observable[Sized]:
        def subscribe(
            observer: abc.ObserverBase[Sized],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_next(item: Sized) -> None:
                self._statistics.submit(len(item))
                observer.on_next(item)

            def on_completed() -> None:
                self._statistics.freeze()
                observer.on_completed()

            return source.subscribe(
                on_next, observer.on_error, on_completed, scheduler=scheduler
            )

        return Observable(subscribe)



from __future__ import annotations

import logging
import time
from typing import Optional, Union

import attr
import m3u8
import requests
import urllib3
from m3u8.model import InitializationSection
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from tenacity import retry, retry_if_exception_type, stop_after_delay, wait_exponential

from blrec.bili.live import Live
from blrec.utils.hash import cksum

from ..exceptions import SegmentDataCorrupted

__all__ = ('SegmentFetcher', 'InitSectionData', 'SegmentData')


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class InitSectionData:
    init_section: InitializationSection
    payload: bytes

    def __len__(self) -> int:
        return len(self.payload)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SegmentData:
    segment: m3u8.Segment
    payload: bytes

    def __len__(self) -> int:
        return len(self.payload)


class SegmentFetcher:
    def __init__(self, live: Live, session: requests.Session) -> None:
        self._live = live
        self._session = session

    def __call__(
        self, source: Observable[m3u8.Segment]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        return self._fetch(source)

    def _fetch(
        self, source: Observable[m3u8.Segment]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        def subscribe(
            observer: abc.ObserverBase[Union[InitSectionData, SegmentData]],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            last_segment: Optional[m3u8.Segment] = None

            def on_next(seg: m3u8.Segment) -> None:
                nonlocal last_segment
                url: str = ''

                try:
                    if hasattr(seg, 'init_section') and (
                        (
                            last_segment is None
                            or seg.init_section != last_segment.init_section
                            or seg.discontinuity
                        )
                    ):
                        url = seg.init_section.absolute_uri
                        data = self._fetch_segment(url)
                        while True:
                            time.sleep(1)
                            if (_data := self._fetch_segment(url)) == data:
                                logger.debug(
                                    'Init section checked: '
                                    f'crc32 of previous data: {cksum(data)}, '
                                    f'crc32 of current data: {cksum(_data)}, '
                                    f'init section url: {url}'
                                )
                                break
                            else:
                                logger.debug(
                                    'Init section corrupted: '
                                    f'crc32 of previous data: {cksum(data)}, '
                                    f'crc32 of current data: {cksum(_data)}, '
                                    f'init section url: {url}'
                                )
                                data = _data
                        observer.on_next(
                            InitSectionData(init_section=seg.init_section, payload=data)
                        )
                    last_segment = seg

                    url = seg.absolute_uri
                    crc32 = seg.title.split('|')[-1]
                    for _ in range(3):
                        data = self._fetch_segment(url)
                        crc32_of_data = cksum(data)
                        if crc32_of_data == crc32:
                            break
                        logger.debug(
                            'Segment data corrupted: '
                            f'correct crc32: {crc32}, '
                            f'crc32 of segment data: {crc32_of_data}, '
                            f'segment url: {url}'
                        )
                    else:
                        raise SegmentDataCorrupted(crc32, crc32_of_data)
                except Exception as exc:
                    logger.warning(f'Failed to fetch segment {url}', exc_info=exc)
                else:
                    observer.on_next(SegmentData(segment=seg, payload=data))

            def dispose() -> None:
                nonlocal disposed
                nonlocal last_segment
                disposed = True
                last_segment = None

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    @retry(
        reraise=True,
        retry=retry_if_exception_type(
            (requests.exceptions.RequestException, urllib3.exceptions.HTTPError)
        ),
        wait=wait_exponential(max=10),
        stop=stop_after_delay(60),
    )
    def _fetch_segment(self, url: str) -> bytes:
        with self._session.get(url, headers=self._live.headers, timeout=10) as response:
            response.raise_for_status()
            return response.content



import os
import logging
from typing import Optional, Tuple

from fastapi import FastAPI, status, Request, Depends
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from starlette.responses import Response
from pydantic import ValidationError
from pkg_resources import resource_filename

from . import security
from .routers import (
    tasks, settings, application, validation, websockets, update
)
from .schemas import ResponseMessage
from ..setting import EnvSettings, Settings
from ..application import Application
from ..exception import NotFoundError, ExistsError, ForbiddenError
from ..path.helpers import file_exists, create_file


logger = logging.getLogger(__name__)


_env_settings = EnvSettings()
_path = os.path.abspath(os.path.expanduser(_env_settings.settings_file))
if not file_exists(_path):
    create_file(_path)
_env_settings.settings_file = _path

_settings = Settings.load(_env_settings.settings_file)
_settings.update_from_env_settings(_env_settings)

app = Application(_settings)

if _env_settings.api_key is None:
    _dependencies = None
else:
    security.api_key = _env_settings.api_key
    _dependencies = [Depends(security.authenticate)]

api = FastAPI(
    title='Bilibili live streaming recorder web API',
    description='Web API to communicate with the backend application',
    version='v1',
    dependencies=_dependencies,
)

api.add_middleware(
    CORSMiddleware,
    allow_origins=[
        'http://localhost:4200',  # angular development
    ],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)


@api.exception_handler(NotFoundError)
async def not_found_error_handler(
    request: Request, exc: NotFoundError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_404_NOT_FOUND,
        content=dict(ResponseMessage(
            code=status.HTTP_404_NOT_FOUND,
            message=str(exc),
        )),
    )


@api.exception_handler(ForbiddenError)
async def forbidden_error_handler(
    request: Request, exc: ForbiddenError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_403_FORBIDDEN,
        content=dict(ResponseMessage(
            code=status.HTTP_403_FORBIDDEN,
            message=str(exc),
        )),
    )


@api.exception_handler(ExistsError)
async def exists_error_handler(
    request: Request, exc: ExistsError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_409_CONFLICT,
        content=dict(ResponseMessage(
            code=status.HTTP_409_CONFLICT,
            message=str(exc),
        )),
    )


@api.exception_handler(ValidationError)
async def validation_error_handler(
    request: Request, exc: ValidationError
) -> JSONResponse:
    return JSONResponse(
        status_code=status.HTTP_406_NOT_ACCEPTABLE,
        content=dict(ResponseMessage(
            code=status.HTTP_406_NOT_ACCEPTABLE,
            message=str(exc),
        )),
    )


@api.on_event('startup')
async def on_startup() -> None:
    await app.launch()


@api.on_event('shutdown')
async def on_shuntdown() -> None:
    _settings.dump()
    await app.exit()


tasks.app = app
settings.app = app
application.app = app
validation.app = app
websockets.app = app
update.app = app
api.include_router(tasks.router)
api.include_router(settings.router)
api.include_router(application.router)
api.include_router(validation.router)
api.include_router(websockets.router)
api.include_router(update.router)


class WebAppFiles(StaticFiles):
    async def lookup_path(
        self, path: str
    ) -> Tuple[str, Optional[os.stat_result]]:
        if path == '404.html':
            path = 'index.html'
        return await super().lookup_path(path)

    def file_response(self, full_path: str, *args, **kwargs) -> Response:  # type: ignore # noqa
        # ignore MIME types from Windows registry
        # workaround for https://github.com/acgnhiki/blrec/issues/12
        response = super().file_response(full_path, *args, **kwargs)
        if full_path.endswith('.js'):
            js_media_type = 'application/javascript'
            if response.media_type != js_media_type:
                response.media_type = js_media_type
                headers = response.headers
                headers['content-type'] = js_media_type
                response.raw_headers = headers.raw
                del response._headers
        return response


directory = resource_filename(__name__, '../data/webapp')
api.mount('/', WebAppFiles(directory=directory, html=True), name='webapp')



import logging
import os
from typing import Optional, Tuple, Union

from reactivex import Observable, Subject, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from blrec.hls.operators.segment_fetcher import InitSectionData, SegmentData

from .playlist_dumper import PlaylistDumper

__all__ = ('SegmentDumper',)

logger = logging.getLogger(__name__)


class SegmentDumper:
    def __init__(self, playlist_dumper: PlaylistDumper) -> None:
        self._playlist_dumper = playlist_dumper
        self._out_dir: str = ''

        def on_next(args: Tuple[str, int]) -> None:
            path, timestamp = args
            self._out_dir = os.path.dirname(path)

        self._playlist_dumper.file_opened.subscribe(on_next)
        self._file_opened: Subject[Tuple[str, int]] = Subject()
        self._file_closed: Subject[str] = Subject()

    @property
    def file_opened(self) -> Observable[Tuple[str, int]]:
        return self._file_opened

    @property
    def file_closed(self) -> Observable[str]:
        return self._file_closed

    def __call__(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        return self._dump(source)

    def _dump(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        def subscribe(
            observer: abc.ObserverBase[Union[InitSectionData, SegmentData]],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            subscription = SerialDisposable()

            def on_next(item: Union[InitSectionData, SegmentData]) -> None:
                if isinstance(item, InitSectionData):
                    uri = item.init_section.uri
                    path = os.path.join(self._out_dir, 'segments', uri)
                else:
                    uri = item.segment.uri
                    name, ext = os.path.splitext(uri)
                    path = os.path.join(self._out_dir, 'segments', name[:-3], uri)
                os.makedirs(os.path.dirname(path), exist_ok=True)
                try:
                    with open(path, 'wb') as file:
                        file.write(item.payload)
                except Exception as e:
                    logger.error(f'Failed to dump segmemt: {repr(e)}')
                    observer.on_error(e)
                else:
                    observer.on_next(item)

            def dispose() -> None:
                pass

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)



import io
import logging
from typing import Callable, Optional, Tuple

from reactivex import Observable, Subject, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ..io import FlvWriter
from ..models import FlvHeader
from .typing import FLVStream, FLVStreamItem

__all__ = ('Dumper',)

logger = logging.getLogger(__name__)


class Dumper:
    def __init__(
        self,
        path_provider: Callable[..., Tuple[str, int]],
        buffer_size: Optional[int] = None,
    ) -> None:
        self.buffer_size = buffer_size or io.DEFAULT_BUFFER_SIZE  # bytes
        self._path_provider = path_provider
        self._file_opened: Subject[Tuple[str, int]] = Subject()
        self._file_closed: Subject[str] = Subject()
        self._size_updates: Subject[int] = Subject()
        self._timestamp_updates: Subject[int] = Subject()
        self._reset()

    def _reset(self) -> None:
        self._path: str = ''
        self._file: Optional[io.BufferedReader] = None
        self._flv_writer: Optional[FlvWriter] = None

    @property
    def path(self) -> str:
        return self._path

    @property
    def file_opened(self) -> Observable[Tuple[str, int]]:
        return self._file_opened

    @property
    def file_closed(self) -> Observable[str]:
        return self._file_closed

    @property
    def size_updates(self) -> Observable[int]:
        return self._size_updates

    @property
    def timestamp_updates(self) -> Observable[int]:
        return self._timestamp_updates

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._dump(source)

    def _open_file(self) -> None:
        self._path, timestamp = self._path_provider()
        self._file = open(self._path, 'wb', buffering=self.buffer_size)  # type: ignore
        logger.debug(f'Opened file: {self._path}')
        self._file_opened.on_next((self._path, timestamp))

    def _close_file(self) -> None:
        if self._file is not None and not self._file.closed:
            self._file.close()
            logger.debug(f'Closed file: {self._path}')
            self._file_closed.on_next(self._path)

    def _dump(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            subscription = SerialDisposable()

            self._close_file()
            self._reset()

            def on_next(item: FLVStreamItem) -> None:
                try:
                    if isinstance(item, FlvHeader):
                        self._close_file()
                        self._open_file()
                        assert self._file is not None
                        self._flv_writer = FlvWriter(self._file)
                        size = self._flv_writer.write_header(item)
                        self._size_updates.on_next(size)
                        self._timestamp_updates.on_next(0)
                    else:
                        if self._flv_writer is not None:
                            size = self._flv_writer.write_tag(item)
                            self._size_updates.on_next(size)
                            self._timestamp_updates.on_next(item.timestamp)

                    observer.on_next(item)
                except Exception as e:
                    self._close_file()
                    self._reset()
                    observer.on_error(e)

            def on_completed() -> None:
                self._close_file()
                self._reset()
                observer.on_completed()

            def on_error(e: Exception) -> None:
                self._close_file()
                self._reset()
                observer.on_error(e)

            def dispose() -> None:
                self._close_file()
                self._reset()

            subscription.disposable = source.subscribe(
                on_next, on_error, on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)



import asyncio
import hashlib
import logging
import os
from abc import ABC
from datetime import datetime
from typing import Any, Dict, List, Mapping, Optional
from urllib.parse import urlencode

import aiohttp
from tenacity import retry, stop_after_delay, wait_exponential

from .exceptions import ApiRequestError
from .typing import JsonResponse, QualityNumber, ResponseData

__all__ = 'AppApi', 'WebApi'


logger = logging.getLogger(__name__)

TRACE_API_REQ = bool(os.environ.get('TRACE_API_REQ'))


class BaseApi(ABC):
    def __init__(
        self, session: aiohttp.ClientSession, headers: Optional[Dict[str, str]] = None
    ):
        self.base_api_urls: List[str] = ['https://api.bilibili.com']
        self.base_live_api_urls: List[str] = ['https://api.live.bilibili.com']
        self.base_play_info_api_urls: List[str] = ['https://api.live.bilibili.com']

        self._session = session
        self.headers = headers or {}
        self.timeout = 10

    @property
    def headers(self) -> Dict[str, str]:
        return self._headers

    @headers.setter
    def headers(self, value: Dict[str, str]) -> None:
        self._headers = {**value}

    @staticmethod
    def _check_response(json_res: JsonResponse) -> None:
        if json_res['code'] != 0:
            raise ApiRequestError(
                json_res['code'], json_res.get('message') or json_res.get('msg') or ''
            )

    @retry(reraise=True, stop=stop_after_delay(5), wait=wait_exponential(0.1))
    async def _get_json_res(self, *args: Any, **kwds: Any) -> JsonResponse:
        kwds = {'timeout': self.timeout, 'headers': self.headers, **kwds}
        async with self._session.get(*args, **kwds) as res:
            if TRACE_API_REQ:
                logger.debug(f'Request info: {res.request_info}')
            try:
                json_res = await res.json()
            except aiohttp.ContentTypeError:
                text_res = await res.text()
                logger.debug(f'Response text: {text_res[:200]}')
                raise
            self._check_response(json_res)
            return json_res

    async def _get_json(
        self, base_urls: List[str], path: str, *args: Any, **kwds: Any
    ) -> JsonResponse:
        if not base_urls:
            raise ValueError('No base urls')
        exception = None
        for base_url in base_urls:
            url = base_url + path
            try:
                return await self._get_json_res(url, *args, **kwds)
            except Exception as exc:
                exception = exc
                if TRACE_API_REQ:
                    logger.debug(f'Failed to get json from {url}', exc_info=exc)
        else:
            assert exception is not None
            raise exception

    async def _get_jsons_concurrently(
        self, base_urls: List[str], path: str, *args: Any, **kwds: Any
    ) -> List[JsonResponse]:
        if not base_urls:
            raise ValueError('No base urls')
        urls = [base_url + path for base_url in base_urls]
        aws = (self._get_json_res(url, *args, **kwds) for url in urls)
        results = await asyncio.gather(*aws, return_exceptions=True)
        exceptions = []
        json_responses = []
        for idx, item in enumerate(results):
            if isinstance(item, Exception):
                if TRACE_API_REQ:
                    logger.debug(f'Failed to get json from {urls[idx]}', exc_info=item)
                exceptions.append(item)
            elif isinstance(item, dict):
                json_responses.append(item)
            else:
                if TRACE_API_REQ:
                    logger.debug(repr(item))
        if not json_responses:
            raise exceptions[0]
        return json_responses


class AppApi(BaseApi):
    # taken from https://github.com/SocialSisterYi/bilibili-API-collect/blob/master/other/API_sign.md  # noqa
    _appkey = '1d8b6e7d45233436'
    _appsec = '560c52ccd288fed045859ed18bffd973'

    _app_headers = {
        'User-Agent': 'Mozilla/5.0 BiliDroid/6.64.0 (bbcallen@gmail.com) os/android model/Unknown mobi_app/android build/6640400 channel/bili innerVer/6640400 osVer/6.0.1 network/2',  # noqa
        'Connection': 'Keep-Alive',
        'Accept-Encoding': 'gzip',
    }

    @property
    def headers(self) -> Dict[str, str]:
        return self._headers

    @headers.setter
    def headers(self, value: Dict[str, str]) -> None:
        self._headers = {**value, **self._app_headers}

    @classmethod
    def signed(cls, params: Mapping[str, Any]) -> Dict[str, Any]:
        if isinstance(params, Mapping):
            params = dict(sorted({**params, 'appkey': cls._appkey}.items()))
        else:
            raise ValueError(type(params))
        query = urlencode(params, doseq=True)
        sign = hashlib.md5((query + cls._appsec).encode()).hexdigest()
        params.update(sign=sign)
        return params

    async def get_room_play_infos(
        self,
        room_id: int,
        qn: QualityNumber = 10000,
        *,
        only_video: bool = False,
        only_audio: bool = False,
    ) -> List[ResponseData]:
        path = '/xlive/app-room/v2/index/getRoomPlayInfo'
        params = self.signed(
            {
                'actionKey': 'appkey',
                'build': '6640400',
                'channel': 'bili',
                'codec': '0,1',  # 0: avc, 1: hevc
                'device': 'android',
                'device_name': 'Unknown',
                'disable_rcmd': '0',
                'dolby': '1',
                'format': '0,1,2',  # 0: flv, 1: ts, 2: fmp4
                'free_type': '0',
                'http': '1',
                'mask': '0',
                'mobi_app': 'android',
                'need_hdr': '0',
                'no_playurl': '0',
                'only_audio': '1' if only_audio else '0',
                'only_video': '1' if only_video else '0',
                'platform': 'android',
                'play_type': '0',
                'protocol': '0,1',
                'qn': qn,
                'room_id': room_id,
                'ts': int(datetime.utcnow().timestamp()),
            }
        )
        json_responses = await self._get_jsons_concurrently(
            self.base_play_info_api_urls, path, params=params
        )
        return [r['data'] for r in json_responses]

    async def get_info_by_room(self, room_id: int) -> ResponseData:
        path = '/xlive/app-room/v1/index/getInfoByRoom'
        params = self.signed(
            {
                'actionKey': 'appkey',
                'build': '6640400',
                'channel': 'bili',
                'device': 'android',
                'mobi_app': 'android',
                'platform': 'android',
                'room_id': room_id,
                'ts': int(datetime.utcnow().timestamp()),
            }
        )
        json_res = await self._get_json(self.base_live_api_urls, path, params=params)
        return json_res['data']

    async def get_user_info(self, uid: int) -> ResponseData:
        base_api_urls = ['https://app.bilibili.com']
        path = '/x/v2/space'
        params = self.signed(
            {
                'build': '6640400',
                'channel': 'bili',
                'mobi_app': 'android',
                'platform': 'android',
                'ts': int(datetime.utcnow().timestamp()),
                'vmid': uid,
            }
        )
        json_res = await self._get_json(base_api_urls, path, params=params)
        return json_res['data']

    async def get_danmu_info(self, room_id: int) -> ResponseData:
        path = '/xlive/app-room/v1/index/getDanmuInfo'
        params = self.signed(
            {
                'actionKey': 'appkey',
                'build': '6640400',
                'channel': 'bili',
                'device': 'android',
                'mobi_app': 'android',
                'platform': 'android',
                'room_id': room_id,
                'ts': int(datetime.utcnow().timestamp()),
            }
        )
        json_res = await self._get_json(self.base_live_api_urls, path, params=params)
        return json_res['data']


class WebApi(BaseApi):
    async def room_init(self, room_id: int) -> ResponseData:
        path = '/room/v1/Room/room_init'
        params = {'id': room_id}
        json_res = await self._get_json(self.base_live_api_urls, path, params=params)
        return json_res['data']

    async def get_room_play_infos(
        self, room_id: int, qn: QualityNumber = 10000
    ) -> List[ResponseData]:
        path = '/xlive/web-room/v2/index/getRoomPlayInfo'
        params = {
            'room_id': room_id,
            'protocol': '0,1',
            'format': '0,1,2',
            'codec': '0,1',
            'qn': qn,
            'platform': 'web',
            'ptype': 8,
        }
        json_responses = await self._get_jsons_concurrently(
            self.base_play_info_api_urls, path, params=params
        )
        return [r['data'] for r in json_responses]

    async def get_info_by_room(self, room_id: int) -> ResponseData:
        path = '/xlive/web-room/v1/index/getInfoByRoom'
        params = {'room_id': room_id}
        json_res = await self._get_json(self.base_live_api_urls, path, params=params)
        return json_res['data']

    async def get_info(self, room_id: int) -> ResponseData:
        path = '/room/v1/Room/get_info'
        params = {'room_id': room_id}
        json_res = await self._get_json(self.base_live_api_urls, path, params=params)
        return json_res['data']

    async def get_timestamp(self) -> int:
        path = '/av/v1/Time/getTimestamp'
        params = {'platform': 'pc'}
        json_res = await self._get_json(self.base_live_api_urls, path, params=params)
        return json_res['data']['timestamp']

    async def get_user_info(self, uid: int) -> ResponseData:
        path = '/x/space/acc/info'
        params = {'mid': uid}
        json_res = await self._get_json(self.base_api_urls, path, params=params)
        return json_res['data']

    async def get_danmu_info(self, room_id: int) -> ResponseData:
        path = '/xlive/web-room/v1/index/getDanmuInfo'
        params = {'id': room_id}
        json_res = await self._get_json(self.base_live_api_urls, path, params=params)
        return json_res['data']



from typing import Literal

# CRITICAL = 50
# FATAL = CRITICAL
# ERROR = 40
# WARNING = 30
# WARN = WARNING
# INFO = 20
# DEBUG = 10
# NOTSET = 0

LOG_LEVEL = Literal[
    'CRITICAL',
    'FATAL',
    'ERROR',
    'WARNING',
    'INFO',
    'DEBUG',
    'NOTSET',  # equivalent to verbose
]



from __future__ import annotations

import io
import logging
from typing import Final, List, Optional, Union

import urllib3
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from tenacity import Retrying, stop_after_delay, wait_fixed
from tenacity.retry import retry_if_not_exception_type

from blrec.bili.live import Live
from blrec.utils.io import wait_for

from ..stream_remuxer import StreamRemuxer
from .segment_fetcher import InitSectionData, SegmentData

__all__ = ('SegmentRemuxer',)


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class SegmentRemuxer:
    _SEGMENT_DATA_CACHE: Final = 10
    _MAX_SEGMENT_DATA_CACHE: Final = 15

    def __init__(self, live: Live) -> None:
        self._live = live
        self._timeout: float = 10
        self._stream_remuxer = StreamRemuxer(live.room_id, remove_filler_data=True)

    def __call__(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[io.RawIOBase]:
        return self._remux(source)

    def _remux(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[io.RawIOBase]:
        def subscribe(
            observer: abc.ObserverBase[io.RawIOBase],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            init_section_data: Optional[bytes] = None
            segment_data_cache: List[bytes] = []
            self._stream_remuxer.stop()

            def reset() -> None:
                nonlocal init_section_data, segment_data_cache
                init_section_data = None
                segment_data_cache = []
                self._stream_remuxer.stop()

            def write(data: bytes) -> int:
                return wait_for(
                    self._stream_remuxer.input.write,
                    args=(data,),
                    timeout=self._timeout,
                )

            def on_next(data: Union[InitSectionData, SegmentData]) -> None:
                nonlocal init_section_data
                nonlocal segment_data_cache

                if isinstance(data, InitSectionData):
                    init_section_data = data.payload
                    segment_data_cache.clear()
                    logger.debug('Stop stream remuxer for init section')
                    self._stream_remuxer.stop()

                if self._stream_remuxer.exception and not self._stream_remuxer.stopped:
                    logger.debug(
                        'Stop stream remuxer due to '
                        + repr(self._stream_remuxer.exception)
                    )
                    self._stream_remuxer.stop()

                try:
                    if self._stream_remuxer.stopped:
                        self._stream_remuxer.start()
                        while True:
                            ready = self._stream_remuxer.wait(timeout=1)
                            if disposed:
                                return
                            if ready:
                                break

                        observer.on_next(RemuxedStream(self._stream_remuxer))

                        if init_section_data:
                            write(init_section_data)
                        if segment_data_cache:
                            for cached_data in segment_data_cache:
                                write(cached_data)
                        if isinstance(data, InitSectionData):
                            return

                    write(data.payload)
                except Exception as e:
                    logger.warning(f'Failed to write data to stream remuxer: {repr(e)}')
                    logger.debug(f'Stop stream remuxer due to {repr(e)}')
                    self._stream_remuxer.stop()
                    if len(segment_data_cache) >= self._MAX_SEGMENT_DATA_CACHE:
                        segment_data_cache = segment_data_cache[
                            -self._MAX_SEGMENT_DATA_CACHE + 1 :
                        ]
                else:
                    if len(segment_data_cache) >= self._SEGMENT_DATA_CACHE:
                        segment_data_cache = segment_data_cache[
                            -self._SEGMENT_DATA_CACHE + 1 :
                        ]

                segment_data_cache.append(data.payload)

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                reset()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)


class CloseRemuxedStream(Exception):
    pass


class RemuxedStream(io.RawIOBase):
    def __init__(
        self, stream_remuxer: StreamRemuxer, *, read_timeout: float = 10
    ) -> None:
        self._stream_remuxer = stream_remuxer
        self._read_timeout = read_timeout
        self._offset: int = 0

    def read(self, size: int = -1) -> bytes:
        if self._stream_remuxer.stopped:
            ready = self._stream_remuxer.wait(timeout=self._read_timeout)
            if not ready:
                msg = f'Stream remuxer not ready in {self._read_timeout} seconds'
                logger.debug(msg)
                raise EOFError(msg)

        try:
            for attempt in Retrying(
                reraise=True,
                retry=retry_if_not_exception_type(TimeoutError),
                wait=wait_fixed(1),
                stop=stop_after_delay(self._read_timeout),
            ):
                with attempt:
                    data = wait_for(
                        self._stream_remuxer.output.read,
                        args=(size,),
                        timeout=self._read_timeout,
                    )
        except Exception as exc:
            logger.warning(f'Failed to read data from stream remuxer: {repr(exc)}')
            self._stream_remuxer.exception = exc
            raise EOFError(exc)
        else:
            assert data is not None
            self._offset += len(data)
            return data

    def tell(self) -> int:
        return self._offset

    def close(self) -> None:
        if self._stream_remuxer.stopped:
            return
        if self._stream_remuxer.exception:
            return
        logger.debug('Close remuxed stream')
        self._stream_remuxer.exception = CloseRemuxedStream()



from io import BytesIO, SEEK_CUR
from typing import cast

import attr

from .struct_io import StructReader, StructWriter
from .io_protocols import RandomIO
from .exceptions import FlvHeaderError, FlvDataError, FlvTagError
from .models import (
    FlvTag,
    TagType,
    FlvHeader,
    FlvTagHeader,
    TAG_HEADER_SIZE,
    AUDIO_TAG_HEADER_SIZE,
    VIDEO_TAG_HEADER_SIZE,

    AudioTag,
    AudioTagHeader,
    SoundFormat,
    SoundRate,
    SoundSize,
    SoundType,
    AACPacketType,

    VideoTag,
    VideoTagHeader,
    CodecID,
    FrameType,
    AVCPacketType,

    ScriptTag,
)


__all__ = 'FlvParser', 'FlvDumper'


class FlvParser:
    def __init__(
        self,
        stream: RandomIO,
        backup_timestamp: bool = False,
        restore_timestamp: bool = False,
    ) -> None:
        self._stream = stream
        self._backup_timestamp = backup_timestamp
        self._restore_timestamp = restore_timestamp
        self._reader = StructReader(stream)

    def parse_header(self) -> FlvHeader:
        signature = self._reader.read(3).decode()
        if signature != 'FLV':
            raise FlvHeaderError(signature)
        version = self._reader.read_ui8()
        type_flag = self._reader.read_ui8()
        data_offset = self._reader.read_ui32()
        return FlvHeader(signature, version, type_flag, data_offset)

    def parse_previous_tag_size(self) -> int:
        return self._reader.read_ui32()

    def parse_tag(self, *, no_body: bool = False) -> FlvTag:
        offset = self._stream.tell()
        tag_header_data = self._reader.read(TAG_HEADER_SIZE)
        tag_header = self.parse_flv_tag_header(tag_header_data)

        if tag_header.data_size <= 0:
            raise FlvTagError('No tag data', tag_header)

        if tag_header.tag_type == TagType.AUDIO:
            header_data = self._reader.read(AUDIO_TAG_HEADER_SIZE)
            body_size = tag_header.data_size - AUDIO_TAG_HEADER_SIZE
            if no_body:
                self._stream.seek(body_size, SEEK_CUR)
                body = None
            else:
                body = self._reader.read(body_size)
            audio_tag_header = self.parse_audio_tag_header(header_data)
            return AudioTag(
                offset=offset,
                **attr.asdict(tag_header),
                **attr.asdict(audio_tag_header),
                body=body,
            )
        elif tag_header.tag_type == TagType.VIDEO:
            header_data = self._reader.read(VIDEO_TAG_HEADER_SIZE)
            body_size = tag_header.data_size - VIDEO_TAG_HEADER_SIZE
            if no_body:
                self._stream.seek(body_size, SEEK_CUR)
                body = None
            else:
                body = self._reader.read(body_size)
            video_tag_header = self.parse_video_tag_header(header_data)
            return VideoTag(
                offset=offset,
                **attr.asdict(tag_header),
                **attr.asdict(video_tag_header),
                body=body,
            )
        elif tag_header.tag_type == TagType.SCRIPT:
            body_size = tag_header.data_size
            if no_body:
                self._stream.seek(body_size, SEEK_CUR)
                body = None
            else:
                body = self._reader.read(body_size)
            return ScriptTag(
                offset=offset,
                **attr.asdict(tag_header),
                body=body,
            )
        else:
            raise FlvDataError(f'Unsupported tag type: {tag_header.tag_type}')

    def parse_flv_tag_header(self, data: bytes) -> FlvTagHeader:
        reader = StructReader(BytesIO(data))

        flag = reader.read_ui8()
        filtered = bool(flag & 0b0010_0000)
        if filtered:
            raise FlvDataError('Unsupported Filtered FLV Tag', data)

        tag_type = TagType(flag & 0b0001_1111)
        data_size = reader.read_ui24()
        timestamp = reader.read_ui24()
        timestamp_extended = reader.read_ui8()
        stream_id = reader.read_ui24()

        if self._backup_timestamp:
            return FlvTagHeader(
                filtered=filtered,
                tag_type=tag_type,
                data_size=data_size,
                timestamp=timestamp_extended << 24 | timestamp,
                stream_id=timestamp,
            )
        elif self._restore_timestamp:
            return FlvTagHeader(
                filtered=filtered,
                tag_type=tag_type,
                data_size=data_size,
                timestamp=stream_id,
                stream_id=stream_id,
            )
        else:
            return FlvTagHeader(
                filtered=filtered,
                tag_type=tag_type,
                data_size=data_size,
                timestamp=timestamp_extended << 24 | timestamp,
                stream_id=stream_id,
            )

    def parse_audio_tag_header(self, data: bytes) -> AudioTagHeader:
        reader = StructReader(BytesIO(data))
        flag = reader.read_ui8()
        sound_format = SoundFormat(flag >> 4)
        if sound_format != SoundFormat.AAC:
            raise FlvDataError(
                f'Unsupported sound format: {sound_format}', data
            )
        sound_rate = SoundRate((flag >> 2) & 0b0000_0011)
        sound_size = SoundSize((flag >> 1) & 0b0000_0001)
        sound_type = SoundType(flag & 0b0000_0001)
        aac_packet_type = AACPacketType(reader.read_ui8())
        return AudioTagHeader(
            sound_format, sound_rate, sound_size, sound_type, aac_packet_type
        )

    def parse_video_tag_header(self, data: bytes) -> VideoTagHeader:
        reader = StructReader(BytesIO(data))
        flag = reader.read_ui8()
        frame_type = FrameType(flag >> 4)
        codec_id = CodecID(flag & 0b0000_1111)
        if codec_id != CodecID.AVC:
            raise FlvDataError(f'Unsupported video codec: {codec_id}', data)
        avc_packet_type = AVCPacketType(reader.read_ui8())
        composition_time = reader.read_ui24()
        return VideoTagHeader(
            frame_type, codec_id, avc_packet_type, composition_time
        )


class FlvDumper:
    def __init__(self, stream: RandomIO) -> None:
        self._stream = stream
        self._writer = StructWriter(stream)

    def dump_header(self, flv_header: FlvHeader) -> None:
        self._writer.write(flv_header.signature.encode())
        self._writer.write_ui8(flv_header.version)
        self._writer.write_ui8(flv_header.type_flag)
        self._writer.write_ui32(flv_header.data_offset)

    def dump_previous_tag_size(self, size: int) -> None:
        self._writer.write_ui32(size)

    def dump_tag(self, tag: FlvTag) -> None:
        if tag.timestamp < 0:
            raise FlvDataError(f'Incorrect timestamp: {tag.timestamp}', tag)

        self.dump_flv_tag_header(tag)

        if tag.is_audio_tag():
            audio_tag = cast(AudioTag, tag)
            self.dump_audio_tag_header(audio_tag)
        elif tag.is_video_tag():
            video_tag = cast(VideoTag, tag)
            self.dump_video_tag_header(video_tag)
        elif tag.is_script_tag():
            pass
        else:
            raise FlvDataError(f'Unsupported tag type: {tag.tag_type}')

        if tag.body:
            self._writer.write(tag.body)
        else:
            self._stream.seek(tag.tag_end_offset)

    def dump_flv_tag_header(self, tag: FlvTag) -> None:
        self._writer.write_ui8((int(tag.filtered) << 5) | tag.tag_type.value)
        self._writer.write_ui24(tag.data_size)
        self._writer.write_ui24(tag.timestamp & 0x00ffffff)
        self._writer.write_ui8(tag.timestamp >> 24)
        self._writer.write_ui24(tag.stream_id)

    def dump_audio_tag_header(self, tag: AudioTag) -> None:
        if tag.sound_format != SoundFormat.AAC:
            raise FlvDataError(
                f'Unsupported sound format: {tag.sound_format}', tag
            )
        self._writer.write_ui8(
            (tag.sound_format.value << 4) |
            (tag.sound_rate.value << 2) |
            (tag.sound_size.value << 1) |
            tag.sound_type.value
        )
        self._writer.write_ui8(tag.aac_packet_type)

    def dump_video_tag_header(self, tag: VideoTag) -> None:
        if tag.codec_id != CodecID.AVC:
            raise FlvDataError(f'Unsupported video codec: {tag.codec_id}', tag)
        self._writer.write_ui8(
            (tag.frame_type.value << 4) | tag.codec_id.value
        )
        self._writer.write_ui8(tag.avc_packet_type.value)
        self._writer.write_ui24(tag.composition_time)



import os
import re
from pathlib import PurePath

__all__ = (
    'cover_path',
    'create_file',
    'danmaku_path',
    'escape_path',
    'extra_metadata_path',
    'ffmpeg_metadata_path',
    'record_metadata_path',
    'file_exists',
    'raw_danmaku_path',
)


def file_exists(path: str) -> bool:
    return os.path.isfile(path)


def create_file(path: str) -> None:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'x'):
        pass


def danmaku_path(video_path: str) -> str:
    return str(PurePath(video_path).with_suffix('.xml'))


def cover_path(video_path: str, ext: str = 'jpg') -> str:
    return str(PurePath(video_path).with_suffix('.' + ext))


def raw_danmaku_path(video_path: str) -> str:
    return str(PurePath(video_path).with_suffix('.jsonl'))


def extra_metadata_path(video_path: str) -> str:
    return video_path + '.meta.json'


def record_metadata_path(video_path: str) -> str:
    return str(PurePath(video_path).with_suffix('.meta.json'))


def ffmpeg_metadata_path(video_path: str) -> str:
    return video_path + '.meta'


def escape_path(path: str) -> str:
    return re.sub(r'[\\/:*?"<>|]', '', path)



from http import HTTPStatus
from typing import Any, Final, Optional

import aiohttp
from tenacity import (
    retry,
    wait_exponential,
    stop_after_delay,
)

from .typing import JsonResponse, Metadata


__all__ = 'PypiApi',


class PypiApi:
    BASE_URL: Final[str] = 'https://pypi.org/pypi'

    def __init__(self, session: aiohttp.ClientSession):
        self._session = session

    @classmethod
    def _make_url(cls, path: str) -> str:
        return cls.BASE_URL + path

    @retry(
        reraise=True,
        stop=stop_after_delay(5),
        wait=wait_exponential(0.1),
    )
    async def _get(self, *args: Any, **kwds: Any) -> Optional[JsonResponse]:
        try:
            async with self._session.get(
                *args, raise_for_status=True, timeout=10, **kwds
            ) as res:
                return await res.json()
        except aiohttp.ClientResponseError as e:
            if e.status == HTTPStatus.NOT_FOUND:
                return None
            else:
                raise

    async def get_project_metadata(
        self, project_name: str
    ) -> Optional[Metadata]:
        url = self._make_url(f'/{project_name}/json')
        return await self._get(url)

    async def get_release_metadata(
        self, project_name: str, version: str
    ) -> Optional[Metadata]:
        url = self._make_url(f'/{project_name}/{version}/json')
        return await self._get(url)



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Dict, Iterator, Optional

import aiohttp
from tenacity import retry, retry_if_exception_type, stop_after_delay, wait_exponential

from ..bili.exceptions import ApiRequestError
from ..exception import NotFoundError, submit_exception
from ..flv.operators import MetaData, StreamProfile
from .models import DanmakuFileDetail, TaskData, TaskParam, VideoFileDetail
from .task import RecordTask

if TYPE_CHECKING:
    from ..setting import SettingsManager

from ..setting import (
    DanmakuSettings,
    BiliApiSettings,
    HeaderSettings,
    OutputSettings,
    PostprocessingSettings,
    RecorderSettings,
    TaskSettings,
)

__all__ = ('RecordTaskManager',)


logger = logging.getLogger(__name__)


class RecordTaskManager:
    def __init__(self, settings_manager: SettingsManager) -> None:
        self._settings_manager = settings_manager
        self._tasks: Dict[int, RecordTask] = {}

    async def load_all_tasks(self) -> None:
        logger.info('Loading all tasks...')

        settings_list = self._settings_manager.get_settings({'tasks'}).tasks
        assert settings_list is not None

        for settings in settings_list:
            try:
                await self.add_task(settings)
            except Exception as e:
                submit_exception(e)

        logger.info('Load all tasks complete')

    async def destroy_all_tasks(self) -> None:
        logger.info('Destroying all tasks...')
        if not self._tasks:
            return
        await asyncio.wait([t.destroy() for t in self._tasks.values() if t.ready])
        self._tasks.clear()
        logger.info('Successfully destroyed all task')

    def has_task(self, room_id: int) -> bool:
        return room_id in self._tasks

    @retry(
        reraise=True,
        retry=retry_if_exception_type(
            (asyncio.TimeoutError, aiohttp.ClientError, ApiRequestError)
        ),
        wait=wait_exponential(max=10),
        stop=stop_after_delay(60),
    )
    async def add_task(self, settings: TaskSettings) -> None:
        logger.info(f'Adding task {settings.room_id}...')

        task = RecordTask(settings.room_id)
        self._tasks[settings.room_id] = task

        try:
            bili_api = self._settings_manager.get_settings({'bili_api'}).bili_api
            assert bili_api is not None
            self.apply_task_bili_api_settings(settings.room_id, bili_api)
            await self._settings_manager.apply_task_header_settings(
                settings.room_id, settings.header, restart_danmaku_client=False
            )
            await task.setup()

            self._settings_manager.apply_task_output_settings(
                settings.room_id, settings.output
            )
            self._settings_manager.apply_task_danmaku_settings(
                settings.room_id, settings.danmaku
            )
            self._settings_manager.apply_task_recorder_settings(
                settings.room_id, settings.recorder
            )
            self._settings_manager.apply_task_postprocessing_settings(
                settings.room_id, settings.postprocessing
            )

            if settings.enable_monitor:
                await task.enable_monitor()
            if settings.enable_recorder:
                await task.enable_recorder()
        except Exception as e:
            logger.error(f'Failed to add task {settings.room_id} due to: {repr(e)}')
            del self._tasks[settings.room_id]
            raise

        logger.info(f'Successfully added task {settings.room_id}')

    async def remove_task(self, room_id: int) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.disable_recorder(force=True)
        await task.disable_monitor()
        await task.destroy()
        del self._tasks[room_id]

    async def remove_all_tasks(self) -> None:
        coros = [self.remove_task(i) for i, t in self._tasks.items() if t.ready]
        if coros:
            await asyncio.wait(coros)

    async def start_task(self, room_id: int) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.update_info()
        await task.enable_monitor()
        await task.enable_recorder()

    async def stop_task(self, room_id: int, force: bool = False) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.disable_recorder(force)
        await task.disable_monitor()

    async def start_all_tasks(self) -> None:
        await self.update_all_task_infos()
        await self.enable_all_task_monitors()
        await self.enable_all_task_recorders()

    async def stop_all_tasks(self, force: bool = False) -> None:
        await self.disable_all_task_recorders(force)
        await self.disable_all_task_monitors()

    async def enable_task_monitor(self, room_id: int) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.enable_monitor()

    async def disable_task_monitor(self, room_id: int) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.disable_monitor()

    async def enable_all_task_monitors(self) -> None:
        coros = [t.enable_monitor() for t in self._tasks.values() if t.ready]
        if coros:
            await asyncio.wait(coros)

    async def disable_all_task_monitors(self) -> None:
        coros = [t.disable_monitor() for t in self._tasks.values() if t.ready]
        if coros:
            await asyncio.wait(coros)

    async def enable_task_recorder(self, room_id: int) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.enable_recorder()

    async def disable_task_recorder(self, room_id: int, force: bool = False) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.disable_recorder(force)

    async def enable_all_task_recorders(self) -> None:
        coros = [t.enable_recorder() for t in self._tasks.values() if t.ready]
        if coros:
            await asyncio.wait(coros)

    async def disable_all_task_recorders(self, force: bool = False) -> None:
        coros = [t.disable_recorder(force) for t in self._tasks.values() if t.ready]
        if coros:
            await asyncio.wait(coros)

    def get_task_data(self, room_id: int) -> TaskData:
        task = self._get_task(room_id, check_ready=True)
        return self._make_task_data(task)

    def get_all_task_data(self) -> Iterator[TaskData]:
        for task in filter(lambda t: t.ready, self._tasks.values()):
            yield self._make_task_data(task)

    def get_task_param(self, room_id: int) -> TaskParam:
        task = self._get_task(room_id, check_ready=True)
        return self._make_task_param(task)

    def get_task_metadata(self, room_id: int) -> Optional[MetaData]:
        task = self._get_task(room_id, check_ready=True)
        return task.metadata

    def get_task_stream_profile(self, room_id: int) -> StreamProfile:
        task = self._get_task(room_id, check_ready=True)
        return task.stream_profile

    def get_task_video_file_details(self, room_id: int) -> Iterator[VideoFileDetail]:
        task = self._get_task(room_id, check_ready=True)
        yield from task.video_file_details

    def get_task_danmaku_file_details(
        self, room_id: int
    ) -> Iterator[DanmakuFileDetail]:
        task = self._get_task(room_id, check_ready=True)
        yield from task.danmaku_file_details

    def can_cut_stream(self, room_id: int) -> bool:
        task = self._get_task(room_id, check_ready=True)
        return task.can_cut_stream()

    def cut_stream(self, room_id: int) -> bool:
        task = self._get_task(room_id, check_ready=True)
        return task.cut_stream()

    async def update_task_info(self, room_id: int) -> None:
        task = self._get_task(room_id, check_ready=True)
        await task.update_info(raise_exception=True)

    async def update_all_task_infos(self) -> None:
        coros = [
            t.update_info(raise_exception=True) for t in self._tasks.values() if t.ready
        ]
        if coros:
            await asyncio.wait(coros)

    def apply_task_bili_api_settings(
        self, room_id: int, settings: BiliApiSettings
    ) -> None:
        task = self._get_task(room_id)
        task.base_api_urls = settings.base_api_urls
        task.base_live_api_urls = settings.base_live_api_urls
        task.base_play_info_api_urls = settings.base_play_info_api_urls

    async def apply_task_header_settings(
        self,
        room_id: int,
        settings: HeaderSettings,
        *,
        restart_danmaku_client: bool = True,
    ) -> None:
        task = self._get_task(room_id)
        # avoid unnecessary updates that will interrupt connections
        if task.user_agent == settings.user_agent and task.cookie == settings.cookie:
            return
        task.user_agent = settings.user_agent
        task.cookie = settings.cookie
        if restart_danmaku_client:
            await task.restart_danmaku_client()

    def apply_task_output_settings(
        self, room_id: int, settings: OutputSettings
    ) -> None:
        task = self._get_task(room_id)
        task.out_dir = settings.out_dir
        task.path_template = settings.path_template
        task.filesize_limit = settings.filesize_limit
        task.duration_limit = settings.duration_limit

    def apply_task_danmaku_settings(
        self, room_id: int, settings: DanmakuSettings
    ) -> None:
        task = self._get_task(room_id)
        task.danmu_uname = settings.danmu_uname
        task.record_gift_send = settings.record_gift_send
        task.record_free_gifts = settings.record_free_gifts
        task.record_guard_buy = settings.record_guard_buy
        task.record_super_chat = settings.record_super_chat
        task.save_raw_danmaku = settings.save_raw_danmaku

    def apply_task_recorder_settings(
        self, room_id: int, settings: RecorderSettings
    ) -> None:
        task = self._get_task(room_id)
        task.stream_format = settings.stream_format
        task.recording_mode = settings.recording_mode
        task.quality_number = settings.quality_number
        task.fmp4_stream_timeout = settings.fmp4_stream_timeout
        task.read_timeout = settings.read_timeout
        task.disconnection_timeout = settings.disconnection_timeout
        task.buffer_size = settings.buffer_size
        task.save_cover = settings.save_cover
        task.cover_save_strategy = settings.cover_save_strategy

    def apply_task_postprocessing_settings(
        self, room_id: int, settings: PostprocessingSettings
    ) -> None:
        task = self._get_task(room_id)
        task.remux_to_mp4 = settings.remux_to_mp4
        task.inject_extra_metadata = settings.inject_extra_metadata
        task.delete_source = settings.delete_source

    def _get_task(self, room_id: int, check_ready: bool = False) -> RecordTask:
        try:
            task = self._tasks[room_id]
        except KeyError:
            raise NotFoundError(f'no task for the room {room_id}')
        else:
            if check_ready and not task.ready:
                raise NotFoundError(f'the task {room_id} is not ready yet')
            return task

    def _make_task_param(self, task: RecordTask) -> TaskParam:
        return TaskParam(
            out_dir=task.out_dir,
            path_template=task.path_template,
            filesize_limit=task.filesize_limit,
            duration_limit=task.duration_limit,
            base_api_urls=task.base_api_urls,
            base_live_api_urls=task.base_live_api_urls,
            base_play_info_api_urls=task.base_play_info_api_urls,
            user_agent=task.user_agent,
            cookie=task.cookie,
            danmu_uname=task.danmu_uname,
            record_gift_send=task.record_gift_send,
            record_free_gifts=task.record_free_gifts,
            record_guard_buy=task.record_guard_buy,
            record_super_chat=task.record_super_chat,
            save_cover=task.save_cover,
            cover_save_strategy=task.cover_save_strategy,
            save_raw_danmaku=task.save_raw_danmaku,
            stream_format=task.stream_format,
            recording_mode=task.recording_mode,
            quality_number=task.quality_number,
            fmp4_stream_timeout=task.fmp4_stream_timeout,
            read_timeout=task.read_timeout,
            disconnection_timeout=task.disconnection_timeout,
            buffer_size=task.buffer_size,
            remux_to_mp4=task.remux_to_mp4,
            inject_extra_metadata=task.inject_extra_metadata,
            delete_source=task.delete_source,
        )

    def _make_task_data(self, task: RecordTask) -> TaskData:
        return TaskData(
            user_info=task.user_info, room_info=task.room_info, task_status=task.status
        )
